from tqdm.autonotebook import tqdm

import logging
logger = logging.getLogger(__name__)

from yayc import Configurator

from .allskyimage import AllSkyImageModel

from .RichardsonLucy import RichardsonLucy
from .RichardsonLucySimple import RichardsonLucySimple

class ImageDeconvolution:
    """
    A class to reconstruct all-sky images from COSI data based on image deconvolution methods.
    """
    model_classes = {"AllSkyImage": AllSkyImageModel}
    deconvolution_algorithm_classes = {"RL": RichardsonLucy, "RLsimple": RichardsonLucySimple}

    def __init__(self):
        self._dataset = None
        self._initial_model = None
        self._mask = None
        self._parameter = None
        self._model_class = None
        self._deconvolution_class = None

    def set_dataset(self, dataset):
        """
        Set dataset

        Parameters
        ----------
        dataset : list of :py:class:`cosipy.image_deconvolution.ImageDeconvolutionDataInterfaceBase` or its subclass
            Each component contaning an event histogram, a background model, a response matrix, and a coordsys_conversion_matrix.
        """

        self._dataset = dataset
        
        logger.debug(f"dataset for image deconvolution was set -> {self._dataset}")

    def set_mask(self, mask):
        """
        Set dataset

        Parameters
        ----------
        mask: :py:class:`histpy.Histogram`
            A mask which will be applied to a model 
        """

        self._mask = mask

    def read_parameterfile(self, parameter_filepath):
        """
        Read parameters from a yaml file.

        Parameters
        ----------
        parameter_filepath : str or pathlib.Path
            Path of parameter file.
        """

        self._parameter = Configurator.open(parameter_filepath)

        logger.debug(f"parameter file for image deconvolution was set -> {parameter_filepath}")

    @property
    def dataset(self):
        """
        Return the dataset.
        """
        return self._dataset

    @property
    def parameter(self):
        """
        Return the registered parameter.
        """
        return self._parameter

    def override_parameter(self, *args):
        """
        Override parameter

        Parameters
        ----------
        *args
            new parameter

        Examples
        --------
        >>> image_deconvolution.override_parameter("deconvolution:parameter_RL:iteration = 30")
        """
        self._parameter.override(args)

    @property
    def initial_model(self):
        """
        Return the initial model.
        """
        if self._initial_model is None:
            logger.warning("Need to initialize model in the image_deconvolution instance!")

        return self._initial_model

    @property
    def mask(self):
        """
        Return the mask.
        """
        return self._mask

    @property
    def results(self):
        """
        Return the results.
        """
        return self._deconvolution.results

    def initialize(self):
        """
        Initialize an initial model and an image deconvolution algorithm.
        It is mandatory to execute this method before running the image deconvolution.
        """

        logger.info("#### Initialization Starts ####")
        
        self.model_initialization()        

        self.register_deconvolution_algorithm()        

        logger.info("#### Initialization Finished ####")

    def model_initialization(self):
        """
        Create an instance of the model class and set initial values of it.

        Returns
        -------
        bool 
            whether the instantiation and initialization are successfully done.
        """
        # set self._model_class
        model_name = self.parameter['model_definition']['class']

        if not model_name in self.model_classes.keys():
            logger.error(f'The model class "{model_name}" does not exist!')
            raise ValueError

        self._model_class = self.model_classes[model_name]

        # instantiate the model class
        logger.info(f"<< Instantiating the model class {model_name} >>")
        parameter_model_property = Configurator(self.parameter['model_definition']['property'])
        self._initial_model = self._model_class.instantiate_from_parameters(parameter_model_property)

        logger.info("---- parameters ----")
        logger.info(parameter_model_property.dump())

        # setting initial values
        logger.info("<< Setting initial values of the created model object >>")
        parameter_model_initialization = Configurator(self.parameter['model_definition']['initialization'])
        self._initial_model.set_values_from_parameters(parameter_model_initialization)

        # applying a mask to the model if needed
        if self.mask is not None:
            self._initial_model = self._initial_model.mask_pixels(self.mask, 0)

        # axes check
        if not self._check_model_response_consistency():
            logger.error("The model axes mismatches with the reponse in the dataset!")
            raise ValueError

        logger.info("---- parameters ----")
        logger.info(parameter_model_initialization.dump())

    def register_deconvolution_algorithm(self):
        """
        Register the deconvolution algorithm

        Returns
        -------
        bool 
            whether the deconvolution algorithm is successfully registered.
        """
        logger.info("<< Registering the deconvolution algorithm >>")
        parameter_deconvolution = Configurator(self.parameter['deconvolution'])

        algorithm_name = parameter_deconvolution['algorithm']
        algorithm_parameter = Configurator(parameter_deconvolution['parameter'])

        if not algorithm_name in self.deconvolution_algorithm_classes.keys():
            logger.error(f'The algorithm "{algorithm_name}" does not exist!')
            raise ValueError

        self._deconvolution_class = self.deconvolution_algorithm_classes[algorithm_name]
        self._deconvolution = self._deconvolution_class(initial_model = self.initial_model, 
                                                        dataset = self.dataset, 
                                                        mask = self.mask, 
                                                        parameter = algorithm_parameter)

        logger.info("---- parameters ----")
        logger.info(parameter_deconvolution.dump()) 

    def run_deconvolution(self):
        """
        Perform the image deconvolution. Make sure that the initialize method has been conducted.
        
        Returns
        -------
        list
            List containing results (reconstructed image, likelihood etc) at each iteration. 
        """
        logger.info("#### Image Deconvolution Starts ####")
       
        logger.info(f"<< Initialization >>")
        self._deconvolution.initialization()
        
        stop_iteration = False
        for i in tqdm(range(self._deconvolution.iteration_max)):
            if stop_iteration:
                break
            stop_iteration = self._deconvolution.iteration()

        logger.info(f"<< Finalization >>")
        self._deconvolution.finalization()

        logger.info("#### Image Deconvolution Finished ####")

    def _check_model_response_consistency(self):
        """
        Check if the model axes is consistent with the dataset

        Returns
        -------
        bool 
            whether the axes of dataset are consistent with the model.
        """
        
        for data in self.dataset:
            if data.model_axes != self.initial_model.axes:
                return False
        return True



import numpy as np
from astropy.coordinates import SkyCoord, Angle
import astropy.units as u
import inspect
from scoords import Attitude, SpacecraftFrame

# Base class for polarization conventions
class PolarizationConvention:

    def __init__(self):
        """
        Base class for polarization conventions
        """

    _registered_conventions = {}
        
    @classmethod
    def register(cls, name):

        name = name.lower()
        
        def _register(convention_class):
            cls._registered_conventions[name] = convention_class
            return convention_class 
        return _register

    @classmethod
    def get_convention(cls, name, *args, **kwargs):

        if inspect.isclass(name):
            if issubclass(name, PolarizationConvention):
                return name(*args, **kwargs)
            else:
                raise TypeError("Class must be subclass of PolarizationConvention")

        if isinstance(name, PolarizationConvention):
            return name

        if not isinstance(name, str):
            raise TypeError("Input must be str, or PolarizationConvention subclass or object")
        
        name = name.lower()

        try:
            return cls._registered_conventions[name](*args, **kwargs)
        except KeyError as e:
            raise Exception(f"No polarization convention by name '{name}'") from e

    @property
    def frame(self):
        """
        Astropy coordinate frame
        """
        return None
        
    def get_basis(self, source_direction: SkyCoord):
        """
        Get the px,py unit vectors that define the polarization plane on 
        this convention. Polarization angle increments from px to py.

        Parameters
        ----------
        source_direction : SkyCoord
            The direction of the source

        Return
        ------
        px,py : SkyCoord
            Polarization angle increaes from px to py. pz is always
            the opposite of the source direction --i.e. in the direction of the
            particle.
        """

        
# Orthographic projection convention
class OrthographicConvention(PolarizationConvention):

    def __init__(self,
                 ref_vector: SkyCoord = None,
                 clockwise: bool = False):
        """
        The local polarization x-axis points towards an arbitrary reference vector, 
        and the polarization angle increasing counter-clockwise when looking 
        at the source.
        
        Parameters
        ----------
        ref_vector : SkyCoord
            Set the reference vector, defaulting to celestial north if not provided 
            (IAU convention)
        clockwise : bool
            Direction of increasing PA, when looking at the source. Default is false 
            --i.e. counter-clockwise when looking outwards.
            
        """
        if ref_vector is None:
            self.ref_vector = SkyCoord(ra=0 * u.deg, dec=90 * u.deg, frame="icrs")
        else:
            self.ref_vector = ref_vector

        self._sign = 1 if clockwise else -1

    def __repr__(self):
        return f"<OrthographicConvention(starting from {self.ref_vector} and {'clockwise' if self.is_clockwise else 'counter-clockwise'} when looking at the source)>"

    @property
    def is_clockwise(self):
        """
        When looking at the source
        """
        return True if self._sign == 1 else False
    
    @property
    def frame(self):
        return self.ref_vector.frame
        
    def get_basis(self, source_direction: SkyCoord):
        # Extract Cartesian coordinates for the source direction.
        pz = self._sign * source_direction.transform_to(self.frame).cartesian.xyz

        # Broadcast reference vector
        ref = np.expand_dims(self.ref_vector.cartesian.xyz,
                             axis = tuple(np.arange(1,pz.ndim, dtype = int)))

        # Get py. Normalize because pz and ref dot not make 90deg angle
        py = np.cross(pz, ref, axisa = 0, axisb = 0, axisc = 0)
        py /= np.linalg.norm(py, axis = 0, keepdims = True)

        # Get px
        px = np.cross(py, pz, axisa = 0, axisb = 0, axisc = 0)
        
        # To SkyCoord
        px = SkyCoord(*px, representation_type='cartesian', frame = self.frame)
        py = SkyCoord(*py, representation_type='cartesian', frame = self.frame)
        
        return px, py


#https://github.com/zoglauer/megalib/blob/1eaad14c51ec52ad1cb2399a7357fe2ca1074f79/src/cosima/src/MCSource.cc#L3452
class MEGAlibRelative(OrthographicConvention):

    def __init__(self, axis, attitude = None):
        """
        Use a polarization vector which is created the following way: 
        Create an initial polarization vector which is orthogonal on the 
        initial flight direction vector of the particle and the given axis vector 
        (e.g. x-axis for RelativeX). This is a simple crossproduct. Then rotate 
        the polarization vector (right-hand-way) around the initial flight 
        direction vector of the particle by the given rotation angle. 
        """

        if not isinstance(axis, str):
            raise TypeError("Axis must be a string. 'x', 'y' or 'z'.")

        axis = axis.lower()
        
        if axis == 'x':
            ref_vector = SkyCoord(lon=0 * u.deg, lat=0 * u.deg,
                                  frame = SpacecraftFrame(attitude = attitude))
        elif axis == 'y':
            ref_vector = SkyCoord(lon=90 * u.deg, lat=0 * u.deg,
                                  frame = SpacecraftFrame(attitude = attitude))
        elif axis == 'z':
            ref_vector = SkyCoord(lon=0 * u.deg, lat=90 * u.deg,
                                  frame = SpacecraftFrame(attitude = attitude))
        else:
            raise ValueError("Axis must be 'x', 'y' or 'z'.")
            
        super().__init__(ref_vector, clockwise = False)
        
    def get_basis(self, source_direction: SkyCoord):

        # The MEGAlib and orthographic definitions are prett much the same, but
        # they differ on the order of the cross products
        
        # In MEGAlib definition
        # pz = -source_direction = particle_direction
        # px = particle_direction x ref_vector  = pz x ref_vector
        # py = pz x px

        # In the base orthographic definition
        # pz = -source_direction = particle_direction
        # px = py x pz
        # py = source_direction x ref_vector = -pz x ref_vector

        # Therefore
        # px = py_base
        # py = -px_base

        # MEGAlib's PA is counter-clockwise when looking at the sourse

        # Flip px <-> py
        py,px = super().get_basis(source_direction)

        # Sign of px
        py = SkyCoord(-py.cartesian,
                      representation_type = 'cartesian',
                      frame = py.frame)
        
        return px,py

@PolarizationConvention.register("RelativeX")    
class MEGAlibRelativeX(MEGAlibRelative):

    def __init__(self, *args, **kwargs):
        super().__init__('x', *args, **kwargs)
        
@PolarizationConvention.register("RelativeY")    
class MEGAlibRelativeY(MEGAlibRelative):

    def __init__(self, *args, **kwargs):
        super().__init__('y', *args, **kwargs)

@PolarizationConvention.register("RelativeZ")    
class MEGAlibRelativeZ(MEGAlibRelative):

    def __init__(self, *args, **kwargs):
        super().__init__('z', *args, **kwargs)

    
# https://lambda.gsfc.nasa.gov/product/about/pol_convention.html
# https://www.iau.org/static/resolutions/IAU1973_French.pdf
@PolarizationConvention.register("IAU")
class IAUPolarizationConvention(OrthographicConvention):

    def __init__(self):
        """
        The following resolution was adopted by Commissions 25 and 40: 
        'RESOLVED, that the frame of reference for the Stokes parameters 
        is that of Right Ascension and Declination with the position 
        angle of electric-vector maximum, e, starting from North and 
        increasing through East.
        """
        super().__init__(ref_vector = SkyCoord(ra=0 * u.deg, dec=90 * u.deg,
                                               frame="icrs"),
                         clockwise = False)
    
    
# Stereographic projection convention
class StereographicConvention(PolarizationConvention):

    def __init__(self,
                 clockwise: bool = False,
                 attitude: Attitude = None):
        """
        Basis vector follow the steregraphic projection lines. Meant to describe
        polarization in spacecraft coordinate by minimizing the number of undefined location withing the field of view.
        
        Near the boresight --i.e. on axis, center of the FoV, north pole-- it is
        similar to 
        ``OrthographicConvention(ref_vector = SkyCoord(lon = 0*u.deg, lat = 0*u.deg, frame = SpacecraftFrame())``
        however, it has a single undefined point on the opposite end --i.e. south pole,
        back of the detector---
        
        
        Parameters
        ----------
        clockwise : bool
            Direction of increasing PA, when looking at the source. Default is false 
            --i.e. counter-clockwise when looking outwards.
        attitude : Attitude
            Spacecraft orientation
        """

        self._attitude = attitude
        
        self._sign = 1 if clockwise else -1

    @property 
    def frame(self):
        return SpacecraftFrame(attitude = self._attitude)
        
    def get_basis(self, source_direction: SkyCoord):
        # Extract Cartesian coordinates for the source direction
        x, y, z = source_direction.cartesian.xyz

        # Calculate the projection of the reference vector in stereographic coordinates
        px_x = 1 - (x**2 - y**2) / (z + 1) ** 2
        px_y = -2 * x * y / (z + 1) ** 2
        px_z = -2 * x / (z + 1)

        # Combine the components into the projection vector px
        px = np.array([px_x, px_y, px_z])

        # Normalize the projection vector
        norm = np.linalg.norm(px, axis=0)
        px /= norm

        # Calculate the perpendicular vector py using the cross product
        py = self._sign*np.cross([x, y, z], px, axis=0)

        # To SkyCoord
        px = SkyCoord(*px, representation_type='cartesian', frame = self.frame)
        py = SkyCoord(*py, representation_type='cartesian', frame = self.frame)
        
        return px, py



from astromodels.functions.function import Function1D, FunctionMeta, ModelAssertionViolation,Function2D
import astromodels.functions.numba_functions as nb_func
from astromodels.utils.angular_distance import angular_distance
from threeML import Band, DiracDelta, Constant, Line, Quadratic, Cubic, Quartic, StepFunction, StepFunctionUpper, Cosine_Prior, Uniform_prior, PhAbs, Gaussian
import astropy.units as astropy_units
from astropy.units import Quantity
from past.utils import old_div
from scipy.special import gammainc, expi
from scipy.interpolate import interp1d
from scipy import integrate
import numpy as np
import math
import astropy.units as u

from histpy import Histogram, Axes, Axis

class Band_Eflux(Function1D, metaclass=FunctionMeta):
    r"""
    description :
        Band model from Band et al., 1993 where the normalization is the flux defined between a and b
    latex : $ A \begin{cases} x^{\alpha} \exp{(-\frac{x}{E0})} & x \leq (\alpha-\beta) E0 \\ x^{\beta} \exp (\beta-\alpha)\left[(\alpha-\beta) E0\right]^{\alpha-\beta} & x>(\alpha-\beta) E0 \end{cases} $
    parameters :
        K :
            desc : Normalization (flux between a and b)
            initial value : 1.e-5
            min : 1e-50
            is_normalization : False
            transformation : log10
        E0 :
            desc : $\frac{xp}{2+\alpha}$ where xp is peak in the x * x * N (nuFnu if x is an energy)
            initial value : 500
            min : 1
            transformation : log10
        alpha :
            desc : low-energy photon index
            initial value : -1.0
            min : -1.5
            max : 3
        beta :
            desc : high-energy photon index
            initial value : -2.0
            min : -5.0
            max : -1.6
        a :
            desc : lower energy integral bound (keV)
            initial value : 10
            min : 0
            fix: yes
        b :
            desc : upper energy integral bound (keV)
            initial value : 1000
            min : 0
            fix: yes
    """
    
    def _set_units(self, x_unit, y_unit):
        # The normalization has the unit of x * y
        self.K.unit = y_unit * x_unit

        # The break point has always the same dimension as the x variable
        self.E0.unit = x_unit

        # alpha and beta are dimensionless
        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.beta.unit = astropy_units.dimensionless_unscaled

        # a and b have the same units of x
        self.a.unit = x_unit
        self.b.unit = x_unit

    def evaluate(self, x, K, E0, alpha, beta, a, b):
        if alpha < beta:
            raise ModelAssertionViolation("Alpha cannot be less than beta")

        if isinstance(x, astropy_units.Quantity):
            alpha_ = alpha.value
            beta_ = beta.value
            K_ = K.value
            E0_ = E0.value
            a_ = a.value
            b_ = b.value
            x_ = x.value

            unit_ = self.y_unit

        else:
            unit_ = 1.0
            alpha_, beta_, K_, E0_, a_, b_, x_ = alpha, beta, K, E0, a, b, x
            
        spectrum_ = Band(alpha=alpha_,
                         beta=beta_,
                         K=1.0,
                         xp=E0_*(2 + alpha_),
                         piv=1.0)
        A_ = K_ / integrate.quad(spectrum_, a_, b_)[0]

        return nb_func.band_eval(x_, A_, alpha_, beta_, E0_, 1.0) * unit_

class SpecFromDat(Function1D, metaclass=FunctionMeta):
        r"""
        description :
            A  spectrum loaded from a dat file
        parameters :
            K :
                desc : Normalization
                initial value : 1.0
                is_normalization : True
                transformation : log10
                min : 1e-30
                max : 1e3
                delta : 0.1
                units: ph/cm2/s
        properties:
            dat:
                desc: the data file to load
                initial value: test.dat
                defer: True
                units: 
                    energy: keV
                    flux: ph/cm2/s/kev
        """            
        def _set_units(self, x_unit, y_unit):
            
            self.K.unit = y_unit

        def evaluate(self, x, K):
            dataFlux = np.genfromtxt(self.dat.value,comments = "#",usecols = (2),skip_footer=1,skip_header=5)
            dataEn = np.genfromtxt(self.dat.value,comments = "#",usecols = (1),skip_footer=1,skip_header=5)
            
            # Calculate the widths of the energy bins
            ewidths = np.diff(dataEn, append=dataEn[-1])

            # Normalize dataFlux using the energy bin widths
            dataFlux = dataFlux  / np.sum(dataFlux * ewidths)
            
            fun = interp1d(dataEn,dataFlux,fill_value=0,bounds_error=False)
            
            if self._x_unit != None:
                dataEn *= self._x_unit

            result = np.zeros(x.shape) * K * 0

            for i in range(len(x)): result[i] += K*fun(x[i])
            return result


class Wide_Asymm_Gaussian_on_sphere(Function2D, metaclass=FunctionMeta):
    r"""
    description :

        A bidimensional Gaussian function on a sphere (in spherical coordinates)

        see https://en.wikipedia.org/wiki/Gaussian_function#Two-dimensional_Gaussian_function

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        a :

            desc : Standard deviation of the Gaussian distribution (major axis)
            initial value : 10
            min : 0
            max : 90

        e :

            desc : Excentricity of Gaussian ellipse, e^2 = 1 - (b/a)^2, where b is the standard deviation of the Gaussian distribution (minor axis)
            initial value : 0.9
            min : 0
            max : 1

        theta :

            desc : inclination of major axis to a line of constant latitude
            initial value : 10.
            min : -90.0
            max : 90.0

    """
    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 and a have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.a.unit = x_unit
        self.e.unit = u.dimensionless_unscaled
        self.theta.unit = u.degree

    def evaluate(self, x, y, lon0, lat0, a, e, theta):

        lon, lat = x, y

        b = a * np.sqrt(1.0 - e**2)

        dX = np.atleast_1d(angular_distance(lon0, lat0, lon, lat0))
        dY = np.atleast_1d(angular_distance(lon0, lat0, lon0, lat))

        dlon = lon - lon0
        if isinstance(dlon, u.Quantity):
            dlon = (dlon.to(u.degree)).value

        idx = np.logical_and(
            np.logical_or(dlon < 0, dlon > 180),
            np.logical_or(dlon > -180, dlon < -360),
        )
        dX[idx] = -dX[idx]

        idx = lat < lat0
        dY[idx] = -dY[idx]

        if isinstance(theta, u.Quantity):
            phi = (theta.to(u.degree)).value + 90.0
        else:
            phi = theta + 90.0

        cos2_phi = np.power(np.cos(phi * np.pi / 180.0), 2)
        sin2_phi = np.power(np.sin(phi * np.pi / 180.0), 2)

        sin_2phi = np.sin(2.0 * phi * np.pi / 180.0)

        A = old_div(cos2_phi, (2.0 * b**2)) + old_div(sin2_phi, (2.0 * a**2))
        
        B = old_div(-sin_2phi, (4.0 * b**2)) + old_div(sin_2phi, (4.0 * a**2))

        C = old_div(sin2_phi, (2.0 * b**2)) + old_div(cos2_phi, (2.0 * a**2))

        E = -A * np.power(dX, 2) + 2.0 * B * dX * dY - C * np.power(dY, 2)

        return np.power(old_div(180, np.pi), 2) * 1.0 / (2 * np.pi * a * b) * np.exp(E)

    def get_boundaries(self):

        # Truncate the gaussian at 2 times the max of sigma allowed

        min_lat = max(-90.0, self.lat0.value - 2 * self.a.max_value)
        max_lat = min(90.0, self.lat0.value + 2 * self.a.max_value)

        max_abs_lat = max(np.absolute(min_lat), np.absolute(max_lat))

        if (
            max_abs_lat > 89.0
            or 2 * self.a.max_value / np.cos(max_abs_lat * np.pi / 180.0) >= 180.0
        ):

            min_lon = 0.0
            max_lon = 360.0

        else:

            min_lon = self.lon0.value - 2 * self.a.max_value / np.cos(
                max_abs_lat * np.pi / 180.0
            )
            max_lon = self.lon0.value + 2 * self.a.max_value / np.cos(
                max_abs_lat * np.pi / 180.0
            )

            if min_lon < 0.0:

                min_lon = min_lon + 360.0

            elif max_lon > 360.0:

                max_lon = max_lon - 360.0

        return (min_lon, max_lon), (min_lat, max_lat)

    def get_total_spatial_integral(self, z=None):
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)        
        



import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u
from astropy.io import fits
from astropy.time import Time, TimeDelta
from astropy.coordinates import SkyCoord, cartesian_to_spherical, Galactic
from mhealpy import HealpixMap
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib import cm, colors
from scipy import interpolate

from scoords import Attitude, SpacecraftFrame
from cosipy.response import FullDetectorResponse

from .scatt_map import SpacecraftAttitudeMap

import logging
logger = logging.getLogger(__name__)

class SpacecraftFile():

    def __init__(self, time, x_pointings = None, y_pointings = None, \
            z_pointings = None, earth_zenith = None, altitude = None,\
            attitude = None, livetime = None, instrument = "COSI", \
            frame = "galactic"):

        """
        Handles the spacecraft orientation. Calculates the dwell time 
        map and point source response over a certain orientation period. 
        Exports the point source response as RMF and ARF files that can be read by XSPEC.
        
        Parameters
        ----------
        Time : astropy.time.Time
            The time stamps for each pointings. Note this is NOT the time duration.
        x_pointings : astropy.coordinates.SkyCoord, optional
            The pointings (galactic system) of the x axis of the local 
            coordinate system attached to the spacecraft (the default 
            is `None`, which implies no input for the x pointings).
        y_pointings : astropy.coordinates.SkyCoord, optional
            The pointings (galactic system) of the y axis of the local 
            coordinate system attached to the spacecraft (the default 
            is `None`, which implies no input for the y pointings).
        z_pointings : astropy.coordinates.SkyCoord, optional
            The pointings (galactic system) of the z axis of the local 
            coordinate system attached to the spacecraft (the default 
            is `None`, which implies no input for the z pointings).
        earth_zenith : astropy.coordinates.SkyCoord, optional
            The pointings (galactic system) of the Earth zenith (the 
            default is `None`, which implies no input for the earth pointings).
	    altitude : array, optional 
            Altitude of the spacecraft in km.
        livetime : array, optional 
            Time in seconds the instrument is live for the corresponding 
            energy bin (using left endpoints so that the last entry in 
            the ori file is 0).
        attitude : numpy.ndarray, optional 
            The attitude of the spacecraft (the default is `None`, 
            which implies no input for the attitude of the spacecraft).
        instrument : str, optional
            The instrument name (the default is "COSI").
        frame : str, optional
            The frame on which the analysis will be based (the default is "galactic").
        """

        # check if the inputs are valid
        # Time
        if isinstance(time, Time):
            self._time = time
        else:
            raise TypeError("The time should be a astropy.time.Time object")

        # Altitude
        if not isinstance(altitude, (type(None))):
            self._altitude = np.array(altitude)

        # livetime
        if not isinstance(livetime, (type(None))):
            self.livetime = np.array(livetime)

        # x pointings
        if isinstance(x_pointings, (SkyCoord, type(None))):
            self.x_pointings = x_pointings
        else:
            raise TypeError("The x_pointing should be a NoneType or SkyCoord object!")

        # y pointings
        if isinstance(y_pointings, (SkyCoord, type(None))):
            self.y_pointings = y_pointings
        else:
            raise TypeError("The y_pointing should be a NoneType or SkyCoord object!")

        # z pointings
        if isinstance(z_pointings, (SkyCoord, type(None))):
            self.z_pointings = z_pointings
        else:
            raise TypeError("The z_pointing should be a NoneType or SkyCoord object!")
	    
	    # earth pointings
        if isinstance(earth_zenith, (SkyCoord, type(None))):
            self.earth_zenith = earth_zenith
        else:
            raise TypeError("The earth_zenith should be a NoneType or SkyCoord object!")    

        # check if the x, y and z pointings are all None (no inputs). If all None, tt will try to read from attitude parameter
        if self.x_pointings is None and self.y_pointings is None and self.z_pointings is None:
            if attitude != None:
                if type(attitude) is Attitude:
                    self.attitude = attitude
                else:
                    raise TypeError("The attitude must be `scoords.attitude.Attitude` object")
            else:
                raise ValueError("Please input the pointings of as least two axes or attitude!")

        else:
            self.attitude = None  # if you have the inputs of x, y and z pointings, the attitude will be overwritten by a None value regardless of the input for the attitude variable.

        self._load_time = self._time.to_value(format = "unix")  # this is not necessary, but just to make sure evething works fine...
        self._x_direction = np.array([x_pointings.l.deg, x_pointings.b.deg]).T  # this is not necessary, but just to make sure evething works fine...
        self._z_direction = np.array([z_pointings.l.deg, z_pointings.b.deg]).T  # this is not necessary, but just to make sure evething works fine...
        self._earth_direction = np.array([earth_zenith.l.deg, earth_zenith.b.deg]).T  # this is not necessary, but just to make sure evething works fine...
      
        self.frame = frame
                       
    @classmethod
    def parse_from_file(cls, file):

        """
        Parses timestamps, axis positions from file and returns to __init__.

        Parameters
        ----------
        file : str
            The file path of the pointings.

        Returns
        -------
        cosipy.spacecraftfile.SpacecraftFile
            The SpacecraftFile object.
        """

        orientation_file = np.loadtxt(file, usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9),delimiter=' ', skiprows=1, comments=("#", "EN"))
        time_stamps = orientation_file[:, 0]
        axis_1 = orientation_file[:, [2, 1]]
        axis_2 = orientation_file[:, [4, 3]]
        axis_3 = orientation_file[:, [7, 6]]
        altitude = np.array(orientation_file[:, 5]) 
        livetime = np.array(orientation_file[:, 8])
        livetime = livetime[:-1] # left end points, so remove last bin. 

        time = Time(time_stamps, format = "unix")
        xpointings = SkyCoord(l = axis_1[:,0]*u.deg, b = axis_1[:,1]*u.deg, frame = "galactic")
        zpointings = SkyCoord(l = axis_2[:,0]*u.deg, b = axis_2[:,1]*u.deg, frame = "galactic")
        earthpointings = SkyCoord(l = axis_3[:,0]*u.deg, b = axis_3[:,1]*u.deg, frame = "galactic")
        
        return cls(time, x_pointings = xpointings, z_pointings = zpointings, earth_zenith = earthpointings, altitude = altitude, livetime=livetime)

    def get_time(self, time_array = None):

        """
        Return the array pf pointing times as a astropy.Time object.

        Parameters
        ----------
        time_array : numpy.ndarray, optional
            The time array (the default is `None`, which implies the time array will be taken from the instance).

        Returns
        -------
        astropy.time.Time
            The time stamps of the orientation.
        """

        if time_array == None:
            self._time = Time(self._load_time, format = "unix")
        else:
            self._time = Time(time_array, format = "unix")

        return self._time

    def get_altitude(self):

        """
        Return the array of Earth altitude.

        

        Returns
        -------
        numpy array
            the Earth altitude.
        """   

        return self._altitude

    def get_time_delta(self, time_array = None):

        """
        Return an array of the time period between neighbouring time points.

        Parameters
        ----------
        time_array : numpy.ndarray, optional
           The time delta array (the default is `None`, which implies the time array will be taken from the instance).

        Returns
        -------
        time_delta : astropy.time.Time
            The time difference between the neighbouring time stamps.
        """

        if time_array == None:
            self._time_delta = np.diff(self._load_time)
        else:
            self._time_delta = np.diff(time_array)

        time_delta = TimeDelta(self._time_delta * u.second)

        return time_delta

    def interpolate_direction(self, trigger, idx, direction):

        """
        Linearly interpolates position at a given time between two timestamps.

        Parameters
        ----------
        trigger : astropy.time.Time
            The time of the event.
        idx : int
            The closest index in the pointing to the trigger time.
        direction : numpy.ndarray
            The pointing axis (x,z).

        Returns
        -------
        numpy.ndarray
            The interpolated positions.
        """

        new_direction_lat = np.interp(trigger.value, self._load_time[idx : idx + 2], direction[idx : idx + 2, 1])
        if (direction[idx, 0] > direction[idx + 1, 0]):
            new_direction_long = np.interp(trigger.value, self._load_time[idx : idx + 2], [direction[idx, 0], 360 + direction[idx + 1, 0]])
            new_direction_long = new_direction_long - 360
        else:
            new_direction_long = np.interp(trigger.value, self._load_time[idx : idx + 2], direction[idx : idx + 2, 0])

        return np.array([new_direction_long, new_direction_lat])

    def source_interval(self, start, stop):

        """
        Returns the SpacecraftFile file class object for the source interval.

        Parameters
        ----------
        start : astropy.time.Time
            The start time of the orientation period.
        stop : astropy.time.Time
            The end time of the orientation period.

        Returns
        -------
        cosipy.spacecraft.SpacecraftFile
        """

        if(start.format != 'unix' or stop.format != 'unix'):
            start = Time(start.unix, format='unix')
            stop = Time(stop.unix, format='unix')

        if(start > stop):
            raise ValueError("start time cannot be after stop time.")

        stop_idx = self._load_time.searchsorted(stop.value)

        if (start.value % 1 == 0):
            start_idx = self._load_time.searchsorted(start.value)
            new_times = self._load_time[start_idx : stop_idx + 1]
            new_x_direction = self._x_direction[start_idx : stop_idx + 1]
            new_z_direction = self._z_direction[start_idx : stop_idx + 1]
            new_earth_direction = self._earth_direction[start_idx : stop_idx + 1]
            new_earth_altitude = self._altitude[start_idx : stop_idx + 1]

        else:
            start_idx = self._load_time.searchsorted(start.value) - 1

            x_direction_start = self.interpolate_direction(start, start_idx, self._x_direction)
            z_direction_start = self.interpolate_direction(start, start_idx, self._z_direction)
            earth_direction_start = self.interpolate_direction(start, start_idx, self._earth_direction)

            new_times = self._load_time[start_idx + 1 : stop_idx + 1]
            new_times = np.insert(new_times, 0, start.value)

            new_x_direction = self._x_direction[start_idx + 1 : stop_idx + 1]
            new_x_direction = np.insert(new_x_direction, 0, x_direction_start, axis = 0)

            new_z_direction = self._z_direction[start_idx + 1 : stop_idx + 1]
            new_z_direction = np.insert(new_z_direction, 0, z_direction_start, axis = 0)
	    
            new_earth_direction = self._earth_direction[start_idx + 1 : stop_idx + 1]
            new_earth_direction = np.insert(new_earth_direction, 0, earth_direction_start, axis = 0)

            # Use linear interpolation to get starting altitude at desired time. 
            f = interpolate.interp1d(self._time.value, self._altitude, kind="linear")
            starting_alt = f(start.value)
            new_earth_altitude = self._altitude[start_idx + 1 : stop_idx + 1]  
            new_earth_altitude = np.insert(new_earth_altitude, 0, starting_alt)


        if (stop.value % 1 != 0):
            stop_idx = self._load_time.searchsorted(stop.value) - 1

            x_direction_stop = self.interpolate_direction(stop, stop_idx, self._x_direction)
            z_direction_stop = self.interpolate_direction(stop, stop_idx, self._z_direction)
            earth_direction_stop = self.interpolate_direction(stop, stop_idx, self._earth_direction)

            new_times = np.delete(new_times, -1)
            new_times = np.append(new_times, stop.value)

            new_x_direction = new_x_direction[:-1]
            new_x_direction = np.append(new_x_direction, [x_direction_stop], axis = 0)

            new_z_direction = new_z_direction[:-1]
            new_z_direction = np.append(new_z_direction, [z_direction_stop], axis = 0)
            
            new_earth_direction = new_earth_direction[:-1]
            new_earth_direction = np.append(new_earth_direction, [earth_direction_stop], axis = 0)
            
            # Use linear interpolation to get starting altitude at desired time.
            f = interpolate.interp1d(self._time.value, self._altitude, kind="linear")
            stop_alt = f(stop.value)
            new_earth_altitude = new_earth_altitude[:-1]
            new_earth_altitude = np.append(new_earth_altitude, [stop_alt])

        time = Time(new_times, format = "unix")
        xpointings = SkyCoord(l = new_x_direction[:,0]*u.deg, b = new_x_direction[:,1]*u.deg, frame = "galactic")
        zpointings = SkyCoord(l = new_z_direction[:,0]*u.deg, b = new_z_direction[:,1]*u.deg, frame = "galactic")
        earthpointings = SkyCoord(l = new_earth_direction[:,0]*u.deg, b = new_earth_direction[:,1]*u.deg, frame = "galactic")
        altitude = new_earth_altitude

        return self.__class__(time, x_pointings = xpointings, z_pointings = zpointings, earth_zenith = earthpointings, altitude =altitude)
      
    def get_attitude(self, x_pointings = None, y_pointings = None, z_pointings = None):

        """
        Converts the x, y and z pointings to the attitude of the telescope.

        Parameters
        ----------
        x_pointings : astropy.coordinates.SkyCoord, optional
            The pointings (galactic system) of the x axis of the local coordinate system attached to the spacecraft (the default is `None`, which implies that the x pointings will be taken from the instance).
        y_pointings : astropy.coordinates.SkyCoord, optional
            The pointings (galactic system) of the y axis of the local coordinate system attached to the spacecraft (the default is `None`, which implies that the y pointings will be taken from the instance).
        z_pointings : astropy.coordinates.SkyCoord, optional
            The pointings (galactic system) of the z axis of the local coordinate system attached to the spacecraft (the default is `None`, which implies that the z pointings will be taken from the instance).

        Returns
        -------
        scoords.attitude.Attitude
            The attitude of the spacecraft.
        """
        if self.attitude is None:
            # the attitude is None, we will calculate from the x, y and z pointings
            if x_pointings is not None:
                self.x_pointings = x_pointings
            if y_pointings is not None:
                self.y_pointings = y_pointings
            if z_pointings is not None:
                self.z_pointings = z_pointings

            list_ = [self.x_pointings, self.y_pointings, self.z_pointings]
            coord_list_of_path = [x for x in list_ if x!=None]  # check how many pointings the user input

            # Check if the user input pointings from at least two axes
            if len(coord_list_of_path) <= 1:
                raise ValueError("You must input pointings of at least two axes")

            # Check if the inputs are SkyCoord objects
            for i in coord_list_of_path:
                if type(i) != SkyCoord:
                    raise ValueError("The coordiates must be a SkyCoord object")

            self.attitude = Attitude.from_axes(x=self.x_pointings,
                                               y=self.y_pointings,
                                               z=self.z_pointings,
                                               frame = self.frame)

        return self.attitude

    def get_target_in_sc_frame(self, target_name, target_coord, attitude = None, quiet = False, save = False):

        """
        Convert the x, y and z pointings of the spacescraft axes to the path of the source in the spacecraft frame.
        Specify the pointings of at least two axes.

        Parameters
        ----------
        target_name : str
            The name of the target object.
        target_coord : astropy.coordinates.SkyCoord
            The coordinates of the target object.
        attitude: scoords.Attitude, optional
            The attitude of the spacecraft (the default is `None`, which implies the attitude will be taken from the instance).
        quiet : bool, default=False
            Setting `True` to stop printing the messages.
        save : bool, default=False
            Setting `True` to save the target coordinates in the spacecraft frame.

        Returns
        -------
        astropy.coordinates.SkyCoord
            The target coordinates in the spacecraft frame.
        """

        if attitude != None:
            self.attitude = attitude
        else:
            self.attitude = self.get_attitude()

        self.target_name = target_name
        if quiet == False:
            logger.info("Now converting to the Spacecraft frame...")
        self.src_path_cartesian = SkyCoord(np.dot(self.attitude.rot.inv().as_matrix(), target_coord.cartesian.xyz.value),
                                           representation_type = 'cartesian',
                                           frame = SpacecraftFrame())

        # The conversion above is in Cartesian frame, so we have to convert them to the spherical one.

        self.src_path_spherical = cartesian_to_spherical(self.src_path_cartesian.x,
                                                         self.src_path_cartesian.y,
                                                         self.src_path_cartesian.z)
        if quiet == False:
            logger.info(f"Conversion completed!")

        # generate the numpy array of l and b to save to a npy file
        l = np.array(self.src_path_spherical[2].deg)  # note that 0 is Quanty, 1 is latitude and 2 is longitude and they are in rad not deg
        b = np.array(self.src_path_spherical[1].deg)
        self.src_path_lb = np.stack((l,b), axis=-1)

        if save == True:
            np.save(self.target_name+"_source_path_in_SC_frame", self.src_path_lb)

        # convert to SkyCoord objects to get the output object of this method
        self.src_path_skycoord = SkyCoord(self.src_path_lb[:,0], self.src_path_lb[:,1], unit = "deg", frame = SpacecraftFrame())

        return self.src_path_skycoord


    def get_dwell_map(self, response, src_path = None, save = False):

        """
        Generates the dwell time map for the source.

        Parameters
        ----------
        response : str or pathlib.Path
            The path to the response file.
        src_path : astropy.coordinates.SkyCoord, optional
            The movement of source in the detector frame (the default is `None`, which implies that the `src_path` will be read from the instance).
        save : bool, default=False
            Set True to save the dwell time map.

        Returns
        -------
        mhealpy.containers.healpix_map.HealpixMap
            The dwell time map.
        """

        # Define the response
        self.response_file = response

        # Define the dts
        self.dts = self.get_time_delta()
        
        # define the target source path in the SC frame
        if src_path is None:
            path = self.src_path_skycoord
        else:
            path = src_path
        # check if the target source path is astropy.Skycoord object
        if type(path) != SkyCoord:
            raise TypeError("The coordinates of the source movement in the Spacecraft frame must be a SkyCoord object")

        if path.shape[0]-1 != self.dts.shape[0]:
            raise ValueError("The dimensions of the dts or source coordinates are not correct. Please check your inputs.")

        with FullDetectorResponse.open(self.response_file) as response:
            self.dwell_map = HealpixMap(base = response,
                                        coordsys = SpacecraftFrame())

        # Get the unique pixels to weight, and sum all the correspondint weights first, so
        # each pixels needs to be called only once.
        # Based on https://stackoverflow.com/questions/23268605/grouping-indices-of-unique-elements-in-numpy
        
        # remove the last value. Effectively a 0th order interpolations
        pixels, weights = self.dwell_map.get_interp_weights(theta = self.src_path_skycoord[:-1])  

        weighted_duration = weights * self.dts.to_value(u.second)[None]

        pixels = pixels.flatten()
        weighted_duration = weighted_duration.flatten()

        pixels_argsort = np.argsort(pixels)

        pixels = pixels[pixels_argsort]
        weighted_duration = weighted_duration[pixels_argsort]

        first_unique = np.concatenate(([True], pixels[1:] != pixels[:-1]))
        
        pixel_unique = pixels[first_unique]

        splits =  np.nonzero(first_unique)[0][1:]
        pixel_durations = [np.sum(weighted_duration[start:stop]) for start,stop in zip(np.append(0,splits), np.append(splits, pixels.size))]
        
        for pix, dur in zip(pixel_unique, pixel_durations):
            self.dwell_map[pix] += dur

        self.dwell_map.to(u.second, update = False, copy = False)
            
        if save == True:
            self.dwell_map.write_map(self.target_name + "_DwellMap.fits", overwrite = True)

        return self.dwell_map

    def get_scatt_map(self,
                       target_coord,
                       nside,
                       scheme = 'ring',
                       coordsys = 'galactic',
                       r_earth = 6378.0,
                       earth_occ = True
                       ):

        """
        Bin the spacecraft attitude history into a 4D histogram that 
        contains the accumulated time the axes of the spacecraft where 
        looking at a given direction. 

        Parameters
        ----------
        target_coord : astropy.coordinates.SkyCoord
            The coordinates of the target object. 
        nside : int
            The nside of the scatt map.
        scheme : str, optional
            The scheme of the scatt map (the default is "ring")
        coordsys : str, optional
            The coordinate system used in the scatt map (the default is "galactic).
        r_earth : float, optional
            Earth radius in km (default is 6378 km).
        earth_occ : bool, optional
            Option to include Earth occultation in scatt map calculation.
            Default is True. 

        Returns
        -------
        h_ori : cosipy.spacecraftfile.scatt_map.SpacecraftAttitudeMap
            The spacecraft attitude map.
        """
        
        # Get orientations
        timestamps = self.get_time()
        attitudes = self.get_attitude()

        # Altitude at each point in the orbit:
        altitude = self._altitude

        # Earth zenith at each point in the orbit:
        earth_zenith = self.earth_zenith

        # Fill (only 2 axes needed to fully define the orientation)
        h_ori = SpacecraftAttitudeMap(nside = nside,
                                      scheme = scheme,
                                      coordsys = coordsys)
        
        x,y,z = attitudes[:-1].as_axes()
       
        # Get max angle based on altitude:
        max_angle = np.pi - np.arcsin(r_earth/(r_earth + altitude))
        max_angle *= (180/np.pi) # angles in degree
        
        # Calculate angle between source direction and Earth zenith
        # for each time stamp:
        src_angle = target_coord.separation(earth_zenith)
        
        # Get pointings that are occulted by Earth:
        earth_occ_index = src_angle.value >= max_angle

        # Define weights and set to 0 if blocked by Earth:
        weight = self.livetime*u.s

        if earth_occ == True:
            weight[earth_occ_index[:-1]] = 0        
        
        # Fill histogram:
        h_ori.fill(x, y, weight = weight)

        return h_ori


    def get_psr_rsp(self, response = None, dwell_map = None, dts = None):

        """
        Generates the point source response based on the response file and dwell time map.
        dts is used to find the exposure time for this observation.

        Parameters
        ----------
        :response : str or pathlib.Path, optional
            The response for the observation (the defaul is `None`, which implies that the `response` will be read from the instance).
        dwell_map : str, optional
            The time dwell map for the source, you can load saved dwell time map using this parameter if you've saved it before (the defaul is `None`, which implies that the `dwell_map` will be read from the instance).
        dts : numpy.ndarray or str, optional
            The elapsed time for each pointing. It must has the same size as the pointings. If you have saved this array, you can load it using this parameter (the defaul is `None`, which implies that the `dts` will be read from the instance).

        Returns
        -------
        Ei_edges : numpy.ndarray
            The edges of the incident energy.
        Ei_lo : numpy.ndarray
            The lower edges of the incident energy.
        Ei_hi : numpy.ndarray
            The upper edges of the incident energy.
        Em_edges : numpy.ndarray
            The edges of the measured energy.
        Em_lo : numpy.ndarray
            The lower edges of the measured energy.
        Em_hi : numpy.ndarray
            The upper edges of the measured energy.
        areas : numpy.ndarray
            The effective area of each energy bin.
        matrix : numpy.ndarray
            The energy dispersion matrix.
        """

        if response == None:
            pass # will use the response defined in the previous steps
        else:
            self.response_file = response

        if dwell_map is None:  # must use is None, or it throws error!
            pass # will use the dwelltime map calculated in the previous steps
        else:
            self.dwell_map = HealpixMap.read_map(dwell_map)

        if dts == None:
            self.dts = self.get_time_delta()
        else:
            self.dts = TimeDelta(dts*u.second)

        with FullDetectorResponse.open(self.response_file) as response:

            # get point source response
            self.psr = response.get_point_source_response(self.dwell_map)

            self.Ei_edges = np.array(response.axes['Ei'].edges)
            self.Ei_lo = np.float32(self.Ei_edges[:-1])  # use float32 to match the requirement of the data type
            self.Ei_hi = np.float32(self.Ei_edges[1:])

            self.Em_edges = np.array(response.axes['Em'].edges)
            self.Em_lo = np.float32(self.Em_edges[:-1])
            self.Em_hi = np.float32(self.Em_edges[1:])

         # get the effective area and matrix
        logger.info("Getting the effective area ...")
        self.areas = np.float32(np.array(self.psr.project('Ei').to_dense().contents))/self.dts.to_value(u.second).sum()
        spectral_response = np.float32(np.array(self.psr.project(['Ei','Em']).to_dense().contents))
        self.matrix = np.float32(np.zeros((self.Ei_lo.size,self.Em_lo.size))) # initate the matrix

        logger.info("Getting the energy redistribution matrix ...")
        for i in np.arange(self.Ei_lo.size):
            new_raw = spectral_response[i,:]/spectral_response[i,:].sum()
            self.matrix[i,:] = new_raw
        self.matrix = self.matrix.T

        return self.Ei_edges, self.Ei_lo, self.Ei_hi, self.Em_edges, self.Em_lo, self.Em_hi, self.areas, self.matrix


    def get_arf(self, out_name = None):

        """
        Converts the point source response to an arf file that can be read by XSPEC.

        Parameters
        ----------
        out_name: str, optional
            The name of the arf file to save. (the default is `None`, which implies that the saving name will be the target name of the instance).
        """

        if out_name == None:
            self.out_name = self.target_name
        else:
            self.out_name = out_name

        # blow write the arf file
        copyright_string="  FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H "

        ## Create PrimaryHDU
        primaryhdu = fits.PrimaryHDU() # create an empty primary HDU
        primaryhdu.header["BITPIX"] = -32 # since it's an empty HDU, I can just change the data type by resetting the BIPTIX value
        primaryhdu.header["COMMENT"] = copyright_string # add comments
        primaryhdu.header # print headers and their values

        col1_energ_lo = fits.Column(name="ENERG_LO", format="E",unit = "keV", array=self.Em_lo)
        col2_energ_hi = fits.Column(name="ENERG_HI", format="E",unit = "keV", array=self.Em_hi)
        col3_specresp = fits.Column(name="SPECRESP", format="E",unit = "cm**2", array=self.areas)
        cols = fits.ColDefs([col1_energ_lo, col2_energ_hi, col3_specresp]) # create a ColDefs (column-definitions) object for all columns
        specresp_bintablehdu = fits.BinTableHDU.from_columns(cols) # create a binary table HDU object

        specresp_bintablehdu.header.comments["TTYPE1"] =  "label for field   1"
        specresp_bintablehdu.header.comments["TFORM1"] =  "data format of field: 4-byte REAL"
        specresp_bintablehdu.header.comments["TUNIT1"] =  "physical unit of field"
        specresp_bintablehdu.header.comments["TTYPE2"] =  "label for field   2"
        specresp_bintablehdu.header.comments["TFORM2"] =  "data format of field: 4-byte REAL"
        specresp_bintablehdu.header.comments["TUNIT2"] =  "physical unit of field"
        specresp_bintablehdu.header.comments["TTYPE3"] =  "label for field   3"
        specresp_bintablehdu.header.comments["TFORM3"] =  "data format of field: 4-byte REAL"
        specresp_bintablehdu.header.comments["TUNIT3"] =  "physical unit of field"

        specresp_bintablehdu.header["EXTNAME"] = ("SPECRESP","name of this binary table extension")
        specresp_bintablehdu.header["TELESCOP"] = ("COSI","mission/satellite name")
        specresp_bintablehdu.header["INSTRUME"] = ("COSI","instrument/detector name")
        specresp_bintablehdu.header["FILTER"] = ("NONE","filter in use")
        specresp_bintablehdu.header["HDUCLAS1"] = ("RESPONSE","dataset relates to spectral response")
        specresp_bintablehdu.header["HDUCLAS2"] = ("SPECRESP","extension contains an ARF")
        specresp_bintablehdu.header["HDUVERS"] = ("1.1.0","version of format")

        new_arfhdus = fits.HDUList([primaryhdu, specresp_bintablehdu])
        new_arfhdus.writeto(f'{self.out_name}.arf', overwrite=True)

        return

    def get_rmf(self, out_name = None):

        """
        Converts the point source response to an rmf file that can be read by XSPEC.

        Parameters
        ----------
        out_name: str, optional
            The name of the arf file to save. (the default is None, which implies that the saving name will be the target name of the instance).
        """

        if out_name == None:
            self.out_name = self.target_name
        else:
            self.out_name = out_name

        # blow write the arf file
        copyright_string="  FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H "

        ## Create PrimaryHDU
        primaryhdu = fits.PrimaryHDU() # create an empty primary HDU
        primaryhdu.header["BITPIX"] = -32 # since it's an empty HDU, I can just change the data type by resetting the BIPTIX value
        primaryhdu.header["COMMENT"] = copyright_string # add comments
        primaryhdu.header # print headers and their values

        ## Create binary table HDU for MATRIX
        ### prepare colums
        energ_lo = []
        energ_hi = []
        n_grp = []
        f_chan = []
        n_chan = []
        matrix = []
        for i in np.arange(len(self.Ei_lo)):
            energ_lo_temp = np.float32(self.Em_lo[i])
            energ_hi_temp = np.float32(self.Ei_hi[i])

            if self.matrix[:,i].sum() != 0:
                nz_matrix_idx = np.nonzero(self.matrix[:,i])[0] # non-zero index for the matrix
                subsets = np.split(nz_matrix_idx, np.where(np.diff(nz_matrix_idx) != 1)[0]+1)
                n_grp_temp = np.int16(len(subsets))
                f_chan_temp = []
                n_chan_temp = []
                matrix_temp = []
                for m in np.arange(n_grp_temp):
                    f_chan_temp += [subsets[m][0]]
                    n_chan_temp += [len(subsets[m])]
                for m in nz_matrix_idx:
                    matrix_temp += [self.matrix[:,i][m]]
                f_chan_temp = np.int16(np.array(f_chan_temp))
                n_chan_temp = np.int16(np.array(n_chan_temp))
                matrix_temp = np.float32(np.array(matrix_temp))
            else:
                n_grp_temp = np.int16(0)
                f_chan_temp = np.int16(np.array([0]))
                n_chan_temp = np.int16(np.array([0]))
                matrix_temp = np.float32(np.array([0]))

            energ_lo.append(energ_lo_temp)
            energ_hi.append(energ_hi_temp)
            n_grp.append(n_grp_temp)
            f_chan.append(f_chan_temp)
            n_chan.append(n_chan_temp)
            matrix.append(matrix_temp)

        col1_energ_lo = fits.Column(name="ENERG_LO", format="E",unit = "keV", array=energ_lo)
        col2_energ_hi = fits.Column(name="ENERG_HI", format="E",unit = "keV", array=energ_hi)
        col3_n_grp = fits.Column(name="N_GRP", format="I", array=n_grp)
        col4_f_chan = fits.Column(name="F_CHAN", format="PI(54)", array=f_chan)
        col5_n_chan = fits.Column(name="N_CHAN", format="PI(54)", array=n_chan)
        col6_n_chan = fits.Column(name="MATRIX", format="PE(161)", array=matrix)
        cols = fits.ColDefs([col1_energ_lo, col2_energ_hi, col3_n_grp, col4_f_chan, col5_n_chan, col6_n_chan]) # create a ColDefs (column-definitions) object for all columns
        matrix_bintablehdu = fits.BinTableHDU.from_columns(cols) # create a binary table HDU object

        matrix_bintablehdu.header.comments["TTYPE1"] = "label for field   1 "
        matrix_bintablehdu.header.comments["TFORM1"] = "data format of field: 4-byte REAL"
        matrix_bintablehdu.header.comments["TUNIT1"] = "physical unit of field"
        matrix_bintablehdu.header.comments["TTYPE2"] = "label for field   2"
        matrix_bintablehdu.header.comments["TFORM2"] = "data format of field: 4-byte REAL"
        matrix_bintablehdu.header.comments["TUNIT2"] = "physical unit of field"
        matrix_bintablehdu.header.comments["TTYPE3"] = "label for field   3 "
        matrix_bintablehdu.header.comments["TFORM3"] = "data format of field: 2-byte INTEGER"
        matrix_bintablehdu.header.comments["TTYPE4"] = "label for field   4"
        matrix_bintablehdu.header.comments["TFORM4"] = "data format of field: variable length array"
        matrix_bintablehdu.header.comments["TTYPE5"] = "label for field   5"
        matrix_bintablehdu.header.comments["TFORM5"] = "data format of field: variable length array"
        matrix_bintablehdu.header.comments["TTYPE6"] = "label for field   6"
        matrix_bintablehdu.header.comments["TFORM6"] = "data format of field: variable length array"

        matrix_bintablehdu.header["EXTNAME"] = ("MATRIX","name of this binary table extension")
        matrix_bintablehdu.header["TELESCOP"] = ("COSI","mission/satellite name")
        matrix_bintablehdu.header["INSTRUME"] = ("COSI","instrument/detector name")
        matrix_bintablehdu.header["FILTER"] = ("NONE","filter in use")
        matrix_bintablehdu.header["CHANTYPE"] = ("PI","total number of detector channels")
        matrix_bintablehdu.header["DETCHANS"] = (len(self.Em_lo),"total number of detector channels")
        matrix_bintablehdu.header["HDUCLASS"] = ("OGIP","format conforms to OGIP standard")
        matrix_bintablehdu.header["HDUCLAS1"] = ("RESPONSE","dataset relates to spectral response")
        matrix_bintablehdu.header["HDUCLAS2"] = ("RSP_MATRIX","dataset is a spectral response matrix")
        matrix_bintablehdu.header["HDUVERS"] = ("1.3.0","version of format")
        matrix_bintablehdu.header["TLMIN4"] = (0,"minimum value legally allowed in column 4")

        ## Create binary table HDU for EBOUNDS
        channels = np.int16(np.arange(len(self.Em_lo)))
        e_min = np.float32(self.Em_lo)
        e_max = np.float32(self.Em_hi)

        col1_channels = fits.Column(name="CHANNEL", format="I", array=channels)
        col2_e_min = fits.Column(name="E_MIN", format="E",unit="keV", array=e_min)
        col3_e_max = fits.Column(name="E_MAX", format="E",unit="keV", array=e_max)
        cols = fits.ColDefs([col1_channels, col2_e_min, col3_e_max])
        ebounds_bintablehdu = fits.BinTableHDU.from_columns(cols)

        ebounds_bintablehdu.header.comments["TTYPE1"] = "label for field   1"
        ebounds_bintablehdu.header.comments["TFORM1"] = "data format of field: 2-byte INTEGER"
        ebounds_bintablehdu.header.comments["TTYPE2"] = "label for field   2"
        ebounds_bintablehdu.header.comments["TFORM2"] = "data format of field: 4-byte REAL"
        ebounds_bintablehdu.header.comments["TUNIT2"] = "physical unit of field"
        ebounds_bintablehdu.header.comments["TTYPE3"] = "label for field   3"
        ebounds_bintablehdu.header.comments["TFORM3"] = "data format of field: 4-byte REAL"
        ebounds_bintablehdu.header.comments["TUNIT3"] = "physical unit of field"

        ebounds_bintablehdu.header["EXTNAME"] = ("EBOUNDS","name of this binary table extension")
        ebounds_bintablehdu.header["TELESCOP"] = ("COSI","mission/satellite")
        ebounds_bintablehdu.header["INSTRUME"] = ("COSI","nstrument/detector name")
        ebounds_bintablehdu.header["FILTER"] = ("NONE","filter in use")
        ebounds_bintablehdu.header["CHANTYPE"] = ("PI","channel type (PHA or PI)")
        ebounds_bintablehdu.header["DETCHANS"] = (len(self.Em_lo),"total number of detector channels")
        ebounds_bintablehdu.header["HDUCLASS"] = ("OGIP","format conforms to OGIP standard")
        ebounds_bintablehdu.header["HDUCLAS1"] = ("RESPONSE","dataset relates to spectral response")
        ebounds_bintablehdu.header["HDUCLAS2"] = ("EBOUNDS","dataset is a spectral response matrix")
        ebounds_bintablehdu.header["HDUVERS"] = ("1.2.0","version of format")

        new_rmfhdus = fits.HDUList([primaryhdu, matrix_bintablehdu,ebounds_bintablehdu])
        new_rmfhdus.writeto(f'{self.out_name}.rmf', overwrite=True)

        return

    def get_pha(self, src_counts, errors, rmf_file = None, arf_file = None, bkg_file = None, exposure_time = None, dts = None, telescope="COSI", instrument="COSI"):

        """
        Generate the pha file that can be read by XSPEC. This file stores the counts info of the source.

        Parameters
        ----------
        src_counts : numpy.ndarray
            The counts in each energy band. If you have src_counts with unit counts/kev/s, you must convert it to counts by multiplying it with exposure time and the energy band width.
        errors : numpy.ndarray
            The error for counts. It has the same unit requirement as src_counts.
        rmf_file : str, optional
            The rmf file name to be written into the pha file (the default is `None`, which implies that it uses the rmf file generate by function `get_rmf`)
        arf_file : str, optional
            The arf file name to be written into the pha file (the default is `None`, which implies that it uses the arf file generate by function `get_arf`)
        bkg_file : str, optional
            The background file name (the default is `None`, which implied the `src_counts` is source counts only).
        exposure_time : float, optional
            The exposure time for this source observation (the default is `None`, which implied that the exposure time will be calculated by `dts`).
        dts : numpy.ndarray, optional
            It's used to calculate the exposure time. It has the same effect as `exposure_time`. If both `exposure_time` and `dts` are given, `dts` will write over the exposure_time (the default is `None`, which implies that the `dts` will be read from the instance).
        telescope : str, optional
            The name of the telecope (the default is "COSI").
        instrument : str, optional
            The instrument name (the default is "COSI").
        """

        self.src_counts = src_counts
        self.errors = errors

        if bkg_file != None:
            self.bkg_file = bkg_file
        else:
            self.bkg_file = "None"
        
        self.bkg_file = bkg_file

        if rmf_file != None:
            self.rmf_file = rmf_file
        else:
            self.rmf_file = f'{self.out_name}.rmf'

        if arf_file != None:
            self.arf_file = arf_file
        else:
            self.arf_file = f'{self.out_name}.arf'

        if exposure_time != None:
            self.exposure_time = exposure_time
        if dts != None:
            self.dts = self.__str_or_array(dts)
            self.exposure_time = self.dts.sum()
        self.telescope = telescope
        self.instrument = instrument
        self.channel_number = len(self.src_counts)

        # define other hardcoded inputs
        copyright_string="  FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H "
        channels = np.arange(self.channel_number)

        # Create PrimaryHDU
        primaryhdu = fits.PrimaryHDU() # create an empty primary HDU
        primaryhdu.header["BITPIX"] = -32 # since it's an empty HDU, I can just change the data type by resetting the BIPTIX value
        primaryhdu.header["COMMENT"] = copyright_string # add comments
        primaryhdu.header["TELESCOP"] = telescope # add telescope keyword valie
        primaryhdu.header["INSTRUME"] = instrument # add instrument keyword valie
        primaryhdu.header # print headers and their values

        # Create binary table HDU
        a1 = np.array(channels,dtype="int32") # I guess I need to convert the dtype to match the format J
        a2 = np.array(self.src_counts,dtype="int64")  # int32 is not enough for counts
        a3 = np.array(self.errors,dtype="int64") # int32 is not enough for errors
        col1 = fits.Column(name="CHANNEL", format="J", array=a1)
        col2 = fits.Column(name="COUNTS", format="K", array=a2,unit="count")
        col3 = fits.Column(name="STAT_ERR", format="K", array=a3,unit="count")
        cols = fits.ColDefs([col1, col2, col3]) # create a ColDefs (column-definitions) object for all columns
        bintablehdu = fits.BinTableHDU.from_columns(cols) # create a binary table HDU object

        #add other BinTableHDU hear keywords,their values, and comments
        bintablehdu.header.comments["TTYPE1"] = "label for field 1"
        bintablehdu.header.comments["TFORM1"] = "data format of field: 32-bit integer"
        bintablehdu.header.comments["TTYPE2"] = "label for field 2"
        bintablehdu.header.comments["TFORM2"] = "data format of field: 32-bit integer"
        bintablehdu.header.comments["TUNIT2"] = "physical unit of field 2"


        bintablehdu.header["EXTNAME"] = ("SPECTRUM","name of this binary table extension")
        bintablehdu.header["TELESCOP"] = (self.telescope,"telescope/mission name")
        bintablehdu.header["INSTRUME"] = (self.instrument,"instrument/detector name")
        bintablehdu.header["FILTER"] = ("NONE","filter type if any")
        bintablehdu.header["EXPOSURE"] = (self.exposure_time,"integration time in seconds")
        bintablehdu.header["BACKFILE"] = (self.bkg_file,"background filename")
        bintablehdu.header["BACKSCAL"] = (1,"background scaling factor")
        bintablehdu.header["CORRFILE"] = ("NONE","associated correction filename")
        bintablehdu.header["CORRSCAL"] = (1,"correction file scaling factor")
        bintablehdu.header["CORRSCAL"] = (1,"correction file scaling factor")
        bintablehdu.header["RESPFILE"] = (self.rmf_file,"associated rmf filename")
        bintablehdu.header["ANCRFILE"] = (self.arf_file,"associated arf filename")
        bintablehdu.header["AREASCAL"] = (1,"area scaling factor")
        bintablehdu.header["STAT_ERR"] = (0,"statistical error specified if any")
        bintablehdu.header["SYS_ERR"] = (0,"systematic error specified if any")
        bintablehdu.header["GROUPING"] = (0,"grouping of the data has been defined if any")
        bintablehdu.header["QUALITY"] = (0,"data quality information specified")
        bintablehdu.header["HDUCLASS"] = ("OGIP","format conforms to OGIP standard")
        bintablehdu.header["HDUCLAS1"] = ("SPECTRUM","PHA dataset")
        bintablehdu.header["HDUVERS"] = ("1.2.1","version of format")
        bintablehdu.header["POISSERR"] = (False,"Poissonian errors to be assumed, T as True")
        bintablehdu.header["CHANTYPE"] = ("PI","channel type (PHA or PI)")
        bintablehdu.header["DETCHANS"] = (self.channel_number,"total number of detector channels")

        new_phahdus = fits.HDUList([primaryhdu, bintablehdu])
        new_phahdus.writeto(f'{self.out_name}.pha', overwrite=True)

        return


    def plot_arf(self, file_name = None, save_name = None, dpi = 300):

        """
        Read the arf fits file, plot and save it.

        Parameters
        ----------
        file_name: str, optional
            The directory if the arf fits file (the default is `None`, which implies the file name will be read from the instance).
        save_name: str, optional
            The name of the saved image of effective area (the default is `None`, which implies the file name will be read from the instance).
        dpi: int, optional
            The dpi of the saved image (the default is 300).
        """

        if file_name != None:
            self.file_name = file_name
        else:
            self.file_name = f'{self.out_name}.arf'

        if save_name != None:
            self.save_name = save_name
        else:
            self.save_name = self.out_name

        self.dpi = dpi

        self.arf = fits.open(self.file_name) # read file

        # SPECRESP HDU
        self.specresp_hdu = self.arf["SPECRESP"]

        self.areas = np.array(self.specresp_hdu.data["SPECRESP"])
        self.Em_lo = np.array(self.specresp_hdu.data["ENERG_LO"])
        self.Em_hi = np.array(self.specresp_hdu.data["ENERG_HI"])

        E_center = (self.Em_lo+self.Em_hi)/2
        E_edges = np.append(self.Em_lo,self.Em_hi[-1])

        fig, ax = plt.subplots()
        ax.hist(E_center,E_edges,weights=self.areas,histtype='step')

        ax.set_title("Effective area")
        ax.set_xlabel("Energy[$keV$]")
        ax.set_ylabel(r"Effective area [$cm^2$]")
        ax.set_xscale("log")
        fig.savefig(f"Effective_area_for_{self.save_name}.png", bbox_inches = "tight", pad_inches=0.1, dpi=self.dpi)
        #fig.show()

        return


    def plot_rmf(self, file_name = None, save_name = None, dpi = 300):

        """
        Read the rmf fits file, plot and save it.

        Parameters
        ----------
        file_name: str, optional
            The directory if the arf fits file (the default is `None`, which implies the file name will be read from the instance).
        save_name: str, optional
            The name of the saved image of effective area (the default is `None`, which implies the file name will be read from the instance).
        dpi: int, optional
            The dpi of the saved image (the default is 300).
        """

        if file_name != None:
            self.file_name = file_name
        else:
            self.file_name = f'{self.out_name}.rmf'

        if save_name != None:
            self.save_name = save_name
        else:
            self.save_name = self.out_name

        self.dpi = dpi

        # Read rmf file
        self.rmf = fits.open(self.file_name) # read file

        # Read the ENOUNDS information
        ebounds_ext = self.rmf["EBOUNDS"]
        channel_low = ebounds_ext.data["E_MIN"] # energy bin lower edges for channels (channels are just incident energy bins)
        channel_high = ebounds_ext.data["E_MAX"] # energy bin higher edges for channels (channels are just incident energy bins)

        # Read the MATRIX extension
        matrix_ext = self.rmf['MATRIX']
        #logger.info(repr(matrix_hdu.header[:60]))
        energy_low = matrix_ext.data["ENERG_LO"] # energy bin lower edges for measured energies
        energy_high = matrix_ext.data["ENERG_HI"] # energy bin higher edges for measured energies
        data = matrix_ext.data

        # Create a 2-d numpy array and store probability data into the redistribution matrix
        rmf_matrix = np.zeros((len(energy_low),len(channel_low))) # create an empty matrix
        for i in np.arange(data.shape[0]): # i is the measured energy index, examine the matrix_ext.data rows by rows
            if data[i][5].sum() == 0: # if the sum of probabilities is zero, then skip since there is no data at all
                pass
            else:
                #measured_energy_index = np.argwhere(energy_low == data[157][0])[0][0]
                f_chan = data[i][3] # get the starting channel of each subsets
                n_chann = data[i][4] # get the number of channels in each subsets
                matrix = data[i][5] # get the probabilities of this row (incident energy)
                indices = []
                for k in f_chan:
                    channels = 0
                    channels = np.arange(k,k + n_chann[np.argwhere(f_chan == k)]).tolist() # generate the cha
                    indices += channels # fappend the channels togeter
                indices = np.array(indices)
                for m in indices:
                    rmf_matrix[i][m] = matrix[np.argwhere(indices == m)[0][0]] # write the probabilities into the empty matrix


        # plot the redistribution matrix
        xcenter = np.divide(energy_low+energy_high,2)
        x_center_coords = np.repeat(xcenter, 10)
        y_center_coords = np.tile(xcenter, 10)
        energy_all_edges = np.append(energy_low,energy_high[-1])
        #bin_edges = np.array([incident_energy_bins,incident_energy_bins]) # doesn't work
        bin_edges = np.vstack((energy_all_edges, energy_all_edges))
        #logger.info(bin_edges)

        self.probability = []
        for i in np.arange(10):
            for j in np.arange(10):
                self.probability.append(rmf_matrix[i][j])
        #logger.info(type(probability))

        plt.hist2d(x=x_center_coords,y=y_center_coords,weights=self.probability,bins=bin_edges, norm=LogNorm())
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel("Incident energy [$keV$]")
        plt.ylabel("Measured energy [$keV$]")
        plt.title("Redistribution matrix")
        #plt.xlim([70,10000])
        #plt.ylim([70,10000])
        plt.colorbar(norm=LogNorm())
        plt.savefig(f"Redistribution_matrix_for_{self.save_name}.png", bbox_inches = "tight", pad_inches=0.1, dpi=300)
        #plt.show()

        return



from cosipy.threeml.COSILike import COSILike

from threeML import DataList, Powerlaw, PointSource, Model, JointLikelihood

import numpy as np

from histpy import Histogram, Axis

from scipy import stats

import matplotlib.pyplot as plt

import astropy.io.fits as fits

import logging
logger = logging.getLogger(__name__)

class TSMap:

    """
    Compute the TS map of using `threeML` package.
    """
    
    def __init__(self, *args, **kwargs):
        pass
    
    def link_model_all_plugins(self, dr, data, bkg, sc_orientation, piv, index, other_plugins=None, norm=1, ra=0, dec=0):

        """
        Load the model and plugins

        Parameters
        ----------
        dr : str
            Path to full detector response.
        data : histpy.Histogram
            Binned data. Note: Eventually this should be a cosipy data class.
        bkg : histpy.Histogram
            Binned background model. Note: Eventually this should be a cosipy data class.
        sc_orientation : cosipy.spacecraftfile.SpacecraftFile
            Contains the information of the orientation: timestamps (astropy.Time) and attitudes (scoord.Attitude) that describe
            the spacecraft for the duration of the data included in the analysis.
        piv : float
            The pivotal energy of the spectrum.
        index : float
            The index of the spectrum.
        other_plugins : threeML.plugins, optional
            The plugins from other instruments.
        norm : int, optional
            The norm of the spectrum model (the default is 1).
        ra : float, optional
            The RA of the source model (the default is 0).
        dec : float, optional
            The Dec of the source model (the default is 0).
        """
        
        # necessary inputs
        self.dr             = dr
        self.data           = data
        self.bkg            = bkg
        self.sc_orientation = sc_orientation
        self.piv            = piv
        self.index          = index
        
        # optional inputs (have default value)
        self.other_plugins  = other_plugins
        self.norm           = norm
        self.ra             = ra
        self.dec            = dec
        
        # instantiate plugin by dr, data, bkg and sc_orientation
        # only COSI plugin for now
        self.instantiate_plugin()
        
        # gather all plugins 
        # only COSI plugin for now
        self.gather_all_plugins()
        
        # create model by Powerlaw and PointSource 
        # Powerlaw needs norm (free parameter), piv and index (free parameter)
        # PointSource needs ra and dec
        self.create_model()
        
        # fix index in further 3ML fitting
        self.fix_index()
        
        # put model and all plugins together
        self.like = JointLikelihood(self.model, self.all_plugins, verbose = False)
        
    def instantiate_plugin(self):

        """
        Instantiate the likelihood plugin.
        """
        
        if self.other_plugins == None:
            self.cosi_plugin = COSILike("cosi",
                                        dr = self.dr,
                                        data = self.data, 
                                        bkg = self.bkg, 
                                        sc_orientation = self.sc_orientation)
        else:
            raise RuntimeError("Only COSI plugin for now")
            
    def gather_all_plugins(self):

        """
        Gather all the plugins togather into a DataList.
        """
        
        if self.other_plugins == None:
            self.all_plugins = DataList(self.cosi_plugin)
        else:
            raise RuntimeError("Only COSI plugin for now")
    
    def create_model(self):

        """
        Create the source model.

        Returns
        -------
        astromodels.core.model.Model
            The source model.

        """
        
        self.spectrum = Powerlaw()
        
        self.spectrum.K.value = self.norm # 1/keV/cm2/s
        self.spectrum.piv.value = self.piv # keV
        self.spectrum.index.value = self.index
        
        self.source = PointSource("source", # The name of the source is arbitrary, but needs to be unique
                                  ra = self.ra, 
                                  dec = self.dec,
                                  spectral_shape = self.spectrum)
        
        self.model = Model(self.source)
        
    def fix_index(self):

        """
        Return the index of the source spectrum.
        """
        
        self.source.spectrum.main.Powerlaw.index.fix = True
    
    def ts_fitting(self):

        """
        Peform the ts fitting.
        """
        
        # collect ts_grid_data, ts_grid_bkg and calculate_ts because sometime we may want to skip fiiting
        self.ts_grid_data()
        self.ts_grid_bkg()
        self.calculate_ts()
        
    # iterate ra and dec to find the best fit of data (time consuming)
    def ts_grid_data(self):

        """
        Perform the ts fitting using the data on the different pixels.
        """
        
        # using rad due to mollweide projection
        self.ra_range  = (-np.pi  , np.pi  ) # rad
        self.dec_range = (-np.pi/2, np.pi/2) # rad
        
        self.log_like = Histogram(
            [Axis(np.linspace(*self.ra_range , 50), label = "ra" ), 
             Axis(np.linspace(*self.dec_range, 25), label = "dec"),]
        )
        
        for i in range(self.log_like.axes['ra'].nbins):
            for j in range(self.log_like.axes['dec'].nbins):
        
                # progress
                logger.info(f"\rra = {i:2d}/{self.log_like.axes['ra'].nbins}   ", end = "")
                logger.info(f"dec = {j:2d}/{self.log_like.axes['dec'].nbins}   ", end = "")
        
                # changing the position parameters
                # converting rad to deg due to ra and dec in 3ML PointSource
                if self.log_like.axes['ra'].centers[i] < 0:
                    self.source.position.ra = (self.log_like.axes['ra'].centers[i] + 2*np.pi) * (180/np.pi) # deg
                else:
                    self.source.position.ra = (self.log_like.axes['ra'].centers[i]) * (180/np.pi) # deg
                self.source.position.dec = self.log_like.axes['dec'].centers[j] * (180/np.pi) # deg
                
                # maximum likelihood
                self.like.fit(quiet=True)
                
                # converting the min (- log likelihood) from 3ML to the max log likelihood for TS 
                self.log_like[i, j] = -self.like._current_minimum 
        
    # iterate ra and dec to find the best fit of bkg
    # only see it as constant for now
    # set the normalization to 0, that is, background-only null-hypothesis
    def ts_grid_bkg(self):

        """
        Perform the ts fitting using the background on the different pixels.
        """
        
        # spectrum.K.value need to be 1e-10 otherwise you will have a migrad error
        self.spectrum.K.value = 1e-10
        
        # maximum likelihood
        self.like.fit(quiet=True)
        
        # converting the min (- log likelihood) from 3ML to the max log likelihood for TS 
        self.log_like0 = -self.like._current_minimum 
        
    # calculate TS by ts_grid_data and ts_grid_bkg
    def calculate_ts(self):

        """
        Calculate the TS by the TS of data and background.
        """
        
        self.ts = 2 * (self.log_like - self.log_like0)

        # getting the maximum
        # note that, in our case, since log_like0 is a constant, max(TS) = 2
        self.argmax = np.unravel_index(np.argmax(self.ts), self.ts.nbins)
        self.ts_max = self.ts[self.argmax]
        
    def print_best_fit(self):

        """
        Print the best fit location.
        """
        
        # report the best fit position
        # converting rad to deg due to ra and dec in 3ML PointSource
        if self.ts.axes['ra'].centers[self.argmax[0]] < 0:
            self.best_ra = (self.ts.axes['ra'].centers[self.argmax[0]] + 2*np.pi) * (180/np.pi) # deg
        else:
            self.best_ra = (self.ts.axes['ra'].centers[self.argmax[0]]) * (180/np.pi) # deg
        self.best_dec = self.ts.axes['dec'].centers[self.argmax[1]] * (180/np.pi) # deg
        logger.info(f"Best fit position: RA = {self.best_ra} deg, Dec = {self.best_dec} deg")
        
        # convert to significance based on Wilk's theorem
        logger.info(f"Expected significance: {stats.norm.isf(stats.chi2.sf(self.ts_max, df = 2)):.1f} sigma")
        
    def save_ts(self, output_file_name):

        """
        Save the TS map.

        Parameters
        ----------
        output_file_name : str
            The path to save the ts map.
        """
        
        # save TS to .h5 file
        self.ts.write(output_file_name, overwrite = True)
    
    def load_ts(self, input_file_name):

        """
        Load a ts map from file.

        Parameters
        ----------
        input_file_name : str
            The path to the saved TS map file.
        """
        
        # load .h5 file to TS
        self.ts = Histogram.open(input_file_name)
        
        # getting the maximum
        self.argmax = np.unravel_index(np.argmax(self.ts), self.ts.nbins)
        self.ts_max = self.ts[self.argmax]
    
    # refit the best fit to check norm
    def refit_best_fit(self):

        """
        Refit the best fit to check norm.
        """
        
        # reset self.spectrum.K.value to self.norm (big initial value)
        self.spectrum.K.value = self.norm
        
        # converting rad to deg due to RA and Dec in 3ML PointSource
        if self.ts.axes['ra'].centers[self.argmax[0]] < 0:
            self.source.position.ra = (self.ts.axes['ra'].centers[self.argmax[0]] + 2*np.pi) * (180/np.pi) # deg
        else:
            self.source.position.ra = (self.ts.axes['ra'].centers[self.argmax[0]]) * (180/np.pi) # deg
        self.source.position.dec = self.ts.axes['dec'].centers[self.argmax[1]] * (180/np.pi) # deg
    
        # maximum likelihood
        self.like.fit()
    
        # display the best fit result 
        self.like.results.display()

    def plot_ts_map(self):

        """
        Plot the TS map.
        """
        
        fig, ax = plt.subplots(figsize=(16, 8), subplot_kw={'projection': 'mollweide'}, dpi=120)
        
        _,plot = self.ts.plot(ax, vmin = 0, colorbar = False, zorder=0)
        
        ax.scatter([self.ts.axes['ra'].centers[self.argmax[0]]],[self.ts.axes['dec'].centers[self.argmax[1]]], label = "Max TS", zorder=3)
        
        ax.scatter([20/180*np.pi],[40/180*np.pi], marker = "x", label = "Injected", zorder=2)
        
        # here we also use Wilk's theorem to find the DeltaTS that corresponse to a 90% containment confidence
        ts_thresh = self.ts_max - stats.chi2.isf(1-.9, df = 2)
        contours = ax.contour(self.ts.axes['ra'].centers, 
                              self.ts.axes['dec'].centers, 
                              self.ts.contents.transpose(), 
                              [ts_thresh], colors = 'red', zorder=1)
        contours.collections[0].set_label("90% cont.")
        
        cbar = fig.colorbar(plot)
        cbar.ax.set_ylabel("TS")
        
        ax.set_xlabel('R.A.', fontsize=15);
        ax.set_ylabel('Dec.', fontsize=15);
        ax.tick_params(axis='x', colors='White')
        ax.legend(fontsize=10)
        
    



import numpy as np

from numpy import log, sqrt, sqrt
from math import factorial

class FastNormFit:
    """
    Perform a fast poisson maximum likelihood ratio fit of the normalization of a 
    source over background.

    The likelihood ratio as a function of the norm is computed as follow

    .. math::
    
        TS(N) = 2 \\sum_i \\left( \\frac{\\log P(d_i; b_i+N e_i)}{\\log P(d_i; b_i)} \\right)

    where :math:`P(d; \lambda)` is the Poisson probability of obtaining :math:`d`
    count where :math:`\lambda` is expected on average; :math:`b` is the estimated
    number of background counts; :math:`N` is the normalization; and :math:`e` 
    is the expected excess -i.e. signal- per normalization unit -i.e. the 
    number of excess counts equals :math:`N`. 

    It can be shown that :math:`TS(N)` has analytic derivative of arbitrary 
    order and that the Newton's method is guaranteed to converge if initialized
    at :math:`N=0`.

    .. note::

        The background is not allowed to float. It is assumed the error on the
        estimation of the background is small compared to the fluctuation of the
        background itself (i.e. :math:`N_{B}/N_{off} \\lll 1`).

    .. note::
    
        Because of the Poisson probability, :math:`TS(N)` is only well-defined
        for :math:`N \geq 1`. By default, underfluctuations are set to 
        :math:`TS(N=0) = 0`. For cases when there is benefit in letting the
        normalization float to negative values, you can use `allow_negative`, 
        but in that case the results are only valid in the Gaussian regime.

    Args:
        max_iter (int): Maximum number of iteration
        conv_frac_tol (float): Convergence stopping condition, expressed as the
            ratio between the size of the last step and the current norm value.
        zero_ts_tol (float): If zero_ts_tol < TS < 0, then TS is set to 0 
            without failed flag status (analytically, TS < 0 should never happen)
        allow_negative (bool): Allow the normalization to float toward 
            negative values
    """    

    def __init__(self,
                 max_iter = 1000,
                 conv_frac_tol = 1e-3,
                 zero_ts_tol = 1e-5,
                 allow_negative = False):

        self.max_iter = max_iter
        self.conv_frac_tol = conv_frac_tol
        self.zero_ts_tol = zero_ts_tol
        self.allow_negative = allow_negative

    @staticmethod
    def ts(data, bkg, unit_excess, norm):
        """
        Get TS for a given normalization. 

        Args:
            data (array): Observed counts
            bkg (array): Background estimation. Same size as data. 
                Every element should be >0
            unit_excess (array): Expected excess per normalization unit. 
                Same size as data.
            norm (float or array): Normalization value

        Return:
            float or array: TS, same shape as norm
        """
        
        data = np.array(data).flatten()
        bkg = np.array(bkg).flatten()
        unit_excess = np.array(unit_excess).flatten()
        
        ts = 0 if np.isscalar(norm) else np.zeros(np.shape(norm))

        for d,b,ue in zip(data, bkg, unit_excess): 

            e = norm*ue

            if d == 0:
                ts -= e
            else:
                ts += d * log((b+e) / b) - e
 
        ts *= 2
     
        return ts

    @staticmethod
    def dts(data, bkg, unit_excess, norm, order=1):
        """
        Get the derivative of TS with respecto to the normalization,
        at given normalization. 

        Args:
            data (array): Observed counts
            bkg (array): Background estimation. Same size as data. 
                Every element should be >0
            unit_excess (array): Expected excess per normalization unit. 
                Same size as data.
            norm (float or array): Normalization value
            order (int): Derivative order

        Return:
            float or array: d^n TS / dN^n, same shape as norm
        """

        if order < 1:
            raise ValueError("Order must be > 1")
        
        data = np.array(data).flatten()
        bkg = np.array(bkg).flatten()
        unit_excess = np.array(unit_excess).flatten()

        dts = 0 if np.isscalar(norm) else np.zeros(np.shape(norm))

        for d,b,ue in zip(data, bkg, unit_excess): 

            e = norm*ue

            if d > 0:
                dts += d * (ue/(b+e))**order

        if order == 1:

            for ue in unit_excess: 
        
                dts -= ue

        dts *= 2 * ((-1)**(order-1)) * factorial(order-1)

        return dts
        
    def solve(self, data, bkg, unit_excess):
        """
        Get the maximum TS, fitted normalization and normalization error 
        (:math:`\Delta TS = 1`)

        .. note::

            The normalization error is obtained from approximating the TS 
            function as as parabola (i.e. valid in the Gaussian regime). TS
            and norm are indeed valid in the Poisson regime.

        Args:
            data (array): Observed counts
            bkg (array): Background estimation. Same size as data. 
                Every element should be >0
            unit_excess (array): Expected excess per normalization unit. 
                Same size as data.

        Return:
            (float, float, float, bool): ts, norm, norm error and status (0 = good) 
        """
        
        dts0 = self.dts(data, bkg, unit_excess, 0)

        if dts0 < 0:
            # Underfluctuation

            # A negative norm is not well defined since we use a
            # Poisson Likelihood. When negative norm is allowed we assume
            # that we are in the Gaussian regime

            ddts0 = self.dts(data, bkg, unit_excess, 0, order = 2)

            if self.allow_negative:

                ts = dts0*dts0/2/ddts0

                norm = -dts0/ddts0

                norm_err = sqrt(-2/ddts0)
            
            else:

                ts = 0
                norm = 0

                if ddts0 == 0:

                    norm_err = -1/dts0
                
                else:

                    norm_err = -(sqrt(dts0*dts0 - 2*ddts0) + dts0) / ddts0

            #Never failed, results where analytical
            failed = False
            
            return (ts, norm, norm_err, failed)

        # Newton's method

        conv = False
        norm = 0
        iteration = 0
        
        while (not conv and iteration < self.max_iter):

            step = (-self.dts(data, bkg, unit_excess, norm, order = 1) /
                    self.dts(data, bkg, unit_excess, norm, order = 2)
)
        
            norm += step

            if step/norm < self.conv_frac_tol:
                conv = True

            iteration += 1

        # One extra step using Halley's method to avoid being biased toward 
            
        f = self.dts(data, bkg, unit_excess, norm,1)
        fp = self.dts(data, bkg, unit_excess, norm,2)
        fpp = self.dts(data, bkg, unit_excess, norm,3)
        norm -= 2 * f * fp / (2 * fp * fp - f * fpp)

        # Compute ts, norm error and checks

        ts = self.ts(data, bkg, unit_excess, norm)

        norm_err = sqrt(-2/self.dts(data, bkg, unit_excess, norm, 2))

        failed = (norm < 0  or iteration == self.max_iter)

        if ts < -self.zero_ts_tol:
            failed = True
        elif ts < 0:
            #Assumed to be a numerical error
            ts = 0

        return (ts, norm, norm_err, failed, iteration)



import numpy as np
import healpy as hp
from tqdm.autonotebook import tqdm
import sparse
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord, cartesian_to_spherical, Galactic

from scoords import Attitude, SpacecraftFrame
from histpy import Histogram, Axes, Axis, HealpixAxis

class CoordsysConversionMatrix(Histogram):
    """
    A class for coordinate conversion matrix (ccm).
    """

    def __init__(self, edges, contents = None, sumw2 = None,
                 labels=None, axis_scale = None, sparse = None, unit = None,
                 binning_method = None):
        
        super().__init__(edges, contents = contents, sumw2 = sumw2,
                         labels = labels, axis_scale = axis_scale, sparse = sparse, unit = unit)

        self.binning_method = binning_method #'Time' or 'ScAtt'

    @classmethod
    def time_binning_ccm(cls, full_detector_response, orientation, time_intervals, nside_model = None, is_nest_model = False):
        """
        Calculate a ccm from a given orientation.
        
        Parameters
        ----------
        full_detector_response : :py:class:`cosipy.response.FullDetectorResponse`
            Response
        orientation : :py:class:`cosipy.spacecraftfile.SpacecraftFile`
            Orientation
        time_intervals : :py:class:`np.array`
            The same format of binned_data.axes['Time'].edges
        nside_model : int or None, default None
            If it is None, it will be the same as the NSIDE in the response.
        is_nest_model : bool, default False
            If scheme of the model map is nested, it should be False while it is rare.

        Returns
        -------
        :py:class:`cosipy.image_deconvolution.CoordsysConversionMatrix`
            Its axes are [ "Time", "lb", "NuLambda" ].
        """

        if nside_model is None:
            nside_model = full_detector_response.nside

        axis_time = Axis(edges = time_intervals, label = "Time")
        axis_model_map = HealpixAxis(nside = nside_model, coordsys = "galactic", label = "lb")
        axis_local_map = full_detector_response.axes["NuLambda"]

        axis_coordsys_conv_matrix = [ axis_time, axis_model_map, axis_local_map ] #Time, lb, NuLambda

        contents = []

        for i_time, [init_time, end_time] in tqdm(enumerate(axis_time.bounds), total = len(axis_time.bounds)):
            ccm_thispix = np.zeros((axis_model_map.nbins, axis_local_map.nbins)) # without unit

            init_time = Time(init_time, format = 'unix')
            end_time = Time(end_time, format = 'unix')
    
            filtered_orientation = orientation.source_interval(init_time, end_time)

            for ipix in range(hp.nside2npix(nside_model)):
                l, b = hp.pix2ang(nside_model, ipix, nest=is_nest_model, lonlat=True)
                pixel_coord = SkyCoord(l, b, unit = "deg", frame = 'galactic')

                pixel_movement = filtered_orientation.get_target_in_sc_frame(target_name = f"pixel_{ipix}_{i_time}",
                                                                             target_coord = pixel_coord,
                                                                             quiet = True,
                                                                             save = False)

                dwell_time_map = filtered_orientation.get_dwell_map(response = full_detector_response.filename,
                                                                    src_path = pixel_movement,
                                                                    save = False)

                ccm_thispix[ipix] = dwell_time_map.data 
                # (HealpixMap).data returns the numpy array without its unit. dwell_time_map.unit is u.s.

            ccm_thispix_sparse = sparse.COO.from_numpy( ccm_thispix.reshape((1, axis_model_map.nbins, axis_local_map.nbins)) )

            contents.append(ccm_thispix_sparse)

        coordsys_conv_matrix = cls(axis_coordsys_conv_matrix, contents = sparse.concatenate(contents), unit = u.s, sparse = True)
        
        coordsys_conv_matrix.binning_method = "Time"

        return coordsys_conv_matrix

    @classmethod
    def spacecraft_attitude_binning_ccm(cls, full_detector_response, exposure_table, nside_model = None, use_averaged_pointing = False):
        """
        Calculate a ccm from a given exposure_table.

        Parameters
        ----------
        full_detector_response : :py:class:`cosipy.response.FullDetectorResponse`
            Response
        exposure_table : :py:class:`cosipy.image_deconvolution.SpacecraftAttitudeExposureTable`
            Scatt exposure table
        nside_model : int or None, default None
            If it is None, it will be the same as the NSIDE in the response.
        use_averaged_pointing : bool, default False
            If it is True, first the averaged Z- and X-pointings are calculated.
            Then the dwell time map is calculated once for ach model pixel and each scatt_binning_index.
            If it is False, the dwell time map is calculated for each attitude in zpointing and xpointing in the exposure table.
            Then the calculated dwell time maps are summed up. 
            In the former case, the computation is fast but may lose the angular resolution. 
            In the latter case, the conversion matrix is more accurate but it takes a long time to calculate it.

        Returns
        -------
        :py:class:`cosipy.image_deconvolution.CoordsysConversionMatrix'
            Its axes are [ "ScAtt", "lb", "NuLambda" ].
        """

        if nside_model is None:
            nside_model = full_detector_response.nside
        is_nest_model = True if exposure_table.scheme == 'nest' else False
        nside_local = full_detector_response.nside
        
        n_scatt_bins = len(exposure_table)

        axis_scatt = Axis(edges = np.arange(n_scatt_bins+1), label = "ScAtt")
        axis_model_map = HealpixAxis(nside = nside_model, coordsys = "galactic", scheme = exposure_table.scheme, label = "lb")
        axis_local_map = full_detector_response.axes["NuLambda"]

        axis_coordsys_conv_matrix = [ axis_scatt, axis_model_map, axis_local_map ] #lb, ScAtt, NuLambda
        
        contents = []

        for i_scatt_bin in tqdm(range(n_scatt_bins)):
            ccm_thispix = np.zeros((axis_model_map.nbins, axis_local_map.nbins)) # without unit

            row = exposure_table.iloc[i_scatt_bin]
        
            scatt_binning_index = row['scatt_binning_index']
            num_pointings = row['num_pointings']
            #healpix_index = row['healpix_index']
            zpointing = row['zpointing']
            xpointing = row['xpointing']
            zpointing_averaged = row['zpointing_averaged']
            xpointing_averaged = row['xpointing_averaged']
            delta_time = row['delta_time']
            exposure = row['exposure']
            
            if use_averaged_pointing:
                z = SkyCoord([zpointing_averaged[0]], [zpointing_averaged[1]], frame="galactic", unit="deg")
                x = SkyCoord([xpointing_averaged[0]], [xpointing_averaged[1]], frame="galactic", unit="deg")
            else:
                z = SkyCoord(zpointing.T[0], zpointing.T[1], frame="galactic", unit="deg")
                x = SkyCoord(xpointing.T[0], xpointing.T[1], frame="galactic", unit="deg")
        
            attitude = Attitude.from_axes(x = x, z = z, frame = 'galactic')

            for ipix in range(hp.nside2npix(nside_model)):
                l, b = hp.pix2ang(nside_model, ipix, nest=is_nest_model, lonlat=True)
                pixel_coord = SkyCoord(l, b, unit = "deg", frame = 'galactic')
            
                src_path_cartesian = SkyCoord(np.dot(attitude.rot.inv().as_matrix(), pixel_coord.cartesian.xyz.value),
                                              representation_type = 'cartesian', frame = SpacecraftFrame())
    
                src_path_spherical = cartesian_to_spherical(src_path_cartesian.x, src_path_cartesian.y, src_path_cartesian.z)
    
                l_scr_path = np.array(src_path_spherical[2].deg)  # note that 0 is Quanty, 1 is latitude and 2 is longitude and they are in rad not deg
                b_scr_path = np.array(src_path_spherical[1].deg)
    
                src_path_skycoord = SkyCoord(l_scr_path, b_scr_path, unit = "deg", frame = SpacecraftFrame())
                            
                pixels, weights = axis_local_map.get_interp_weights(src_path_skycoord)
                
                if use_averaged_pointing:
                    weights = weights * exposure
                else:
                    weights = weights * delta_time

                hist, bins = np.histogram(pixels, bins = axis_local_map.edges, weights = weights)
                
                ccm_thispix[ipix] = hist

            ccm_thispix_sparse = sparse.COO.from_numpy( ccm_thispix.reshape((1, axis_model_map.nbins, axis_local_map.nbins)) )

            contents.append(ccm_thispix_sparse)

        coordsys_conv_matrix = cls(axis_coordsys_conv_matrix, contents = sparse.concatenate(contents), unit = u.s, sparse = True)

        coordsys_conv_matrix.binning_method = 'ScAtt'
        
        return coordsys_conv_matrix
    
    @classmethod
    def open(cls, filename, name = 'hist'):
        """
        Open a ccm from a file.

        Parameters
        ----------
        filename : str
            Path to file.
        name : str, default 'hist'
            Name of group where the histogram was saved.

        Returns
        -------
        :py:class:`cosipy.image_deconvolution.CoordsysConversionMatrix'
            Its axes are [ "lb", "Time" or "ScAtt", "NuLambda" ].
        """

        new = super().open(filename, name)

        new = cls(new.axes, contents = new.contents, sumw2 = new.contents, unit = new.unit) 

        new.binning_method = new.axes.labels[0] # 'Time' or 'ScAtt'

        return new

# TODO
#   def calc_exposure_map(self, full_detector_response): #once the response file format is fixed, I will implement this function



# Imports:
import sys
import numpy as np
import h5py
from histpy import Histogram, HealpixAxis, Axis
from scoords import SpacecraftFrame, Attitude
from mhealpy import HealpixMap, HealpixBase
import healpy as hp
import pandas as pd
import matplotlib.pyplot as plt
from cosipy.data_io import UnBinnedData
import logging
import astropy.units as u
from astropy.coordinates import SkyCoord
logger = logging.getLogger(__name__)


class BinnedData(UnBinnedData):
    """Handles binned data."""

    def get_binned_data(self, unbinned_data=None, output_name=None, \
            make_binning_plots=False, show_plots=False, \
            psichi_binning="galactic", event_range=None):

        """Bin the data using histpy and mhealpy.
        
        Parameters
        ----------
        unbinned_data : str, optional
            Name of unbinned data file to use. Input file is either 
            .fits or .hdf5 as specified in the unbinned_output 
            parameter in inputs.yaml.     
        output_name : str, optional
            Prefix of output file.
        make_binning_plots : bool, optional
            Option to make basic plots of the binning (default is False).
        show_plots : bool, optional
            Option to show plots (default is False).
        psichi_binning : str, optional
            'galactic' for binning psichi in Galactic coordinates, or 
            'local' for binning in local coordinates. Default is Galactic. 
        event_range : list of integers, optional 
            min and max event to use for the binning. 
        
        Returns
        -------
        binned_data : histpy:Histogram
            Data is binned in four axes: time, measured energy, 
            Compton scattering angle (phi), and scattering direction 
            (PsiChi).

        Note
        ----
        This method constructs the instance attribute, binned_data, 
        but it does not explicitly return it.
        """
        
        # Log message:
        logger.info("binning data...")

        # Option to read in unbinned data file:
        if unbinned_data:
            self.cosi_dataset = self.get_dict(unbinned_data)

        # Get time bins:
        min_time = self.tmin
        max_time = self.tmax
        if type(self.time_bins).__name__ in ['int','float']:
            # Get time bins: 
            delta_t = max_time - min_time
            num_bins = round(delta_t / self.time_bins)
            new_bin_size = delta_t / num_bins
            if self.time_bins != new_bin_size:
                logger.info("Note: time bins must be equally spaced between min and max time.")
                logger.info("Using time bin size [s]: " + str(new_bin_size))
            time_bin_edges = np.linspace(min_time,max_time,num_bins+1)

        if type(self.time_bins).__name__ == 'list':
            # Check that bins correspond to min and max time:
            if (self.time_bins[0] > min_time) | (self.time_bins[-1] < max_time):
                logger.error("ERROR: Time bins do not cover the full selected data range!")
                sys.exit()
            time_bin_edges = np.array(self.time_bins)

        # Get energy bins:
        energy_bin_edges = np.array(self.energy_bins)

        # Get phi bins:
        number_phi_bins = int(180./self.phi_pix_size)
        phi_bin_edges = np.linspace(0,180,number_phi_bins+1)

        # Define psichi axis and data for binning:
        if psichi_binning == 'galactic':
            psichi_axis = HealpixAxis(nside = self.nside, 
                    scheme = self.scheme, coordsys = 'galactic', label='PsiChi')
            coords = SkyCoord(l=self.cosi_dataset['Chi galactic']*u.deg, 
                    b=self.cosi_dataset['Psi galactic']*u.deg, frame = 'galactic')
        if psichi_binning == 'local':
            psichi_axis = HealpixAxis(nside = self.nside, 
                    scheme = self.scheme, coordsys = SpacecraftFrame(), label='PsiChi')
            coords = SkyCoord(lon=self.cosi_dataset['Chi local']*u.rad, 
                    lat=((np.pi/2.0) - self.cosi_dataset['Psi local'])*u.rad, 
                    frame = SpacecraftFrame())

        # Initialize histogram:
        self.binned_data = Histogram([Axis(time_bin_edges*u.s, label='Time'), 
            Axis(energy_bin_edges*u.keV, label='Em'), 
            Axis(phi_bin_edges*u.deg, label='Phi'), 
            psichi_axis], 
            sparse=True)
         
        # Fill histogram:
        if event_range == None:
            self.binned_data.fill(self.cosi_dataset['TimeTags']*u.s, 
                    self.cosi_dataset['Energies']*u.keV, 
                    np.rad2deg(self.cosi_dataset['Phi'])*u.deg, 
                    coords)
        if event_range != None:
            low = int(event_range[0])
            high = int(event_range[1])
            self.binned_data.fill(self.cosi_dataset['TimeTags'][low:high]*u.s, 
                    self.cosi_dataset['Energies'][low:high]*u.keV, 
                    np.rad2deg(self.cosi_dataset['Phi'][low:high])*u.deg, 
                    coords[low:high])

        # Save binned data to hdf5 file:
        if output_name != None:
            self.binned_data.write('%s.hdf5' %output_name, overwrite=True)

        # Get binning information:
        self.get_binning_info()

        # Plot the binned data:
        if make_binning_plots == True:
            self.plot_binned_data(show_plots=show_plots)  
            self.plot_psichi_map(show_plots=show_plots)

        return

    def load_binned_data_from_hdf5(self,binned_data):

        """Loads binned histogram from hdf5 file.
        
        Parameters
        ----------
        binned_data : str
            Name of binned data file to load. 

        Returns
        -------
        binned_data : histpy:Histogram
            Data is binned in four axes: time, measured energy, 
            Compton scattering angle (phi), and scattering direction 
            (PsiChi).

        Note
        ----
        This method sets the instance attribute, binned_data, 
        but it does not explicitly return it.
        """
        
        self.binned_data = Histogram.open(binned_data)    

        return

    def get_binning_info(self, binned_data=None):

        """Get binning information from Histpy histogram.
        
        Parameters
        ----------
        binned_data : str
            Name of binned data hdf5 file to use.
        """
        
        # Option to read in binned data from hdf5 file:
        if binned_data:
            self.load_binned_data_from_hdf5(binned_data)
       
        # Print units of axes:
        for each in self.binned_data.axes:
            logger.info(each.label + " unit: " + str(each.unit))

        # Get time binning information:
        self.time_hist = self.binned_data.project('Time').contents.todense()
        self.num_time_bins = self.binned_data.axes['Time'].nbins
        self.time_bin_centers = self.binned_data.axes['Time'].centers
        self.time_bin_edges = self.binned_data.axes['Time'].edges
        self.time_bin_widths = self.binned_data.axes['Time'].widths
        self.total_time = self.time_bin_edges[-1] - self.time_bin_edges[0]

        # Get energy binning information:
        self.energy_hist = self.binned_data.project('Em').contents.todense()
        self.num_energy_bins = self.binned_data.axes['Em'].nbins
        self.energy_bin_centers = self.binned_data.axes['Em'].centers
        self.energy_bin_edges = self.binned_data.axes['Em'].edges
        self.energy_bin_widths = self.binned_data.axes['Em'].widths

        # Get Phi binning information:
        self.phi_hist = self.binned_data.project('Phi').contents.todense()
        self.num_phi_bins = self.binned_data.axes['Phi'].nbins
        self.phi_bin_centers = self.binned_data.axes['Phi'].centers
        self.phi_bin_edges = self.binned_data.axes['Phi'].edges
        self.phi_bin_widths = self.binned_data.axes['Phi'].widths

        # Get PsiChi binning information:
        self.psichi_hist = self.binned_data.project('PsiChi').contents.todense()
        self.num_psichi_bins = self.binned_data.axes['PsiChi'].nbins
        self.psichi_bin_centers = self.binned_data.axes['PsiChi'].centers
        self.psichi_bin_edges = self.binned_data.axes['PsiChi'].edges
        self.psichi_bin_widths = self.binned_data.axes['PsiChi'].widths
         
        return

    def plot_binned_data(self, binned_data=None, show_plots=True):

        """Plot binnned data for all axes.
        
        Parameters
        ----------
        binned_data : histpy:Histogram, optional
            Name of binned histogram to use. 
        show_plots : bool, optional
            Option to show plots (default is True).
        """
        
        # Option to read in binned data from hdf5 file:
        if binned_data:
            self.load_binned_data_from_hdf5(binned_data)
    
        # Define plot dictionaries:
        time_energy_plot = {"projection":["Time","Em"],"xlabel":"Time [s]",\
                "ylabel":"Em [keV]","savefig":"2d_time_energy.png"}
        time_plot = {"projection":"Time","xlabel":"Time [s]",\
                "ylabel":"Counts","savefig":"time_binning.pdf"}
        energy_plot = {"projection":"Em","xlabel":"Em [keV]",\
                "ylabel":"Counts","savefig":"energy_binning.pdf"}
        phi_plot = {"projection":"Phi","xlabel":"Phi [deg]",\
                "ylabel":"Counts","savefig":"phi_binning.pdf"}
        psichi_plot = {"projection":"PsiChi",\
                "xlabel":"PsiChi [HEALPix ring pixel number]",\
                "ylabel":"Counts","savefig":"psichi_binning.pdf"}
        
        # Make plots:
        plot_list = [time_energy_plot,time_plot,energy_plot,phi_plot,psichi_plot]
        for each in plot_list:
            self.binned_data.project(each["projection"]).plot()
            plt.xlabel(each["xlabel"],fontsize=12)
            plt.ylabel(each["ylabel"], fontsize=12)
            plt.savefig(each["savefig"])
            if show_plots == True:
                plt.show()
            plt.close() 
 
        return

    def plot_psichi_map(self, show_plots=True):
        
        """
        Plot psichi healpix map.

        Parameters
        ----------
        show_plots : bool, optional
            Option to show plot (default is True).
        """

        logger.info("plotting psichi in Galactic coordinates...")
        plot, ax = self.binned_data.project('PsiChi').plot(ax_kw = {'coord':'G'})
        ax.get_figure().set_figwidth(4)
        ax.get_figure().set_figheight(3)
        plt.title("PsiChi Binning (counts)")
        plt.savefig("psichi_default.png",bbox_inches='tight')
        if show_plots == True:
            plt.show()
        plt.close()

        return

    def plot_psichi_map_slices(self, Em, phi, output, binned_data=None, \
            coords=None, show_plots=True):

        """Plot psichi map in slices of Em and phi.
        
        Parameters
        ----------
        Em : int 
            Bin of energy slice.
        phi : int 
            Bin of phi slice.
        output : str
            Prefix of output plot. 
        binned_data : histpy:Histogram, optional
            Name of binned histogram to use. 
        coords : list, optional
            Coordinates of source position. Galactic longitude and 
            latidude for Galactic coordinates. Azimuthal and latitude 
            for local coordinates. 
        show_plots : bool, optional
            Option to show plots (default is True).
        """

        # Option to read in binned data from hdf5 file:
        if binned_data:
            self.load_binned_data_from_hdf5(binned_data)

        # Make healpix map with binned data slice:
        h = self.binned_data.project('Em', 'Phi', 'PsiChi').slice[{'Em':Em, 'Phi':phi}].project('PsiChi')
        m = HealpixMap(base = HealpixBase(npix = h.nbins), data = h.contents.todense())
        
        # Plot standard view:
        plot,ax = m.plot('mollview')
        if coords:
            ax.scatter(coords[0], coords[1], s=9, transform=ax.get_transform('world'), color = 'red')
        ax.coords.grid(True, color='grey', ls='dotted')
        ax.get_figure().set_figwidth(6)
        ax.get_figure().set_figheight(3)
        plt.savefig("%s.pdf" %output,bbox_inches='tight')
        if show_plots == True:
            plt.show()
        plt.close()

        # Plot rotated view:
        if coords:
            plot,ax = m.plot('orthview', ax_kw = {'rot':[coords[0],coords[1],0]})
            ax.scatter(coords[0], coords[1], s=9, transform=ax.get_transform('world'), color = 'red')
            ax.coords.grid(True, color='grey', ls='dotted')
            ax.get_figure().set_figwidth(6)
            ax.get_figure().set_figheight(3)
            plt.savefig("%s_rotated.pdf" %output,bbox_inches='tight')
            if show_plots == True:
                plt.show()
            plt.close()

        return

    def make_basic_plot(self, x, y, plt_scale='loglog', output_name=None,\
        x_error=[], plot_kwargs={}, fig_kwargs={}, show_plots=True):

        """Make a basic plot.

        Parameters
        ----------
        x : list or array
            x-axis data to be plotted
        y : list or array 
            y-axis data to be plotted
        plt_scale : str, optional
            scale of axes: loglog, semilogx, or semilogy.
            Default is loglog.
        x_error : list or array, optional 
            x error bars
        output_name : str, optional
            Prefix of saved figure (default is None).
        plot_kwargs : dict, optional
            Pass any kwargs to plt.plot().
        fig_kwargs : dict, optional 
            Pass any kwargs to plt.gca().set().
        show_plots : bool
            Wether or not to show plot (default is True).
        """
        
        # Setup figure:
        ax = plt.gca()

        # Main plot:
        if plt_scale == "loglog":
            plt.loglog(x, y, **plot_kwargs)
        if plt_scale == "semilogx":
            plt.semilogx(x, y, **plot_kwargs)
        if plt_scale == "semilogy":
            plt.semilogy(x, y, **plot_kwargs)

        # Include x error bars:
        if len(x_error) != 0:
            
            # Remove label if defined:
            if "label" in plot_kwargs.keys():
                plot_kwargs["label"] = "_nolabel_"

            plt.errorbar(x, y, xerr=x_error, **plot_kwargs)

        # axes and labels:
        plt.grid(color="grey",alpha=0.3,ls="--")
        ax.set(**fig_kwargs)
        if "label" in plot_kwargs.keys():
            plt.legend(loc=1,frameon=True)
        
        # Save and show:
        if output_name != None:
            plt.savefig("%s.pdf" %output_name)
        if show_plots == True:
            plt.show()
        plt.close()

        return
    
    def get_raw_spectrum(self, binned_data=None, time_rate=False, output_name=None, show_plots=False):

        """Calculates raw spectrum of binned data, plots, and writes to file. 
        
        Parameters
        ----------
        binned_data : str, optional
            Name of binnned hdf5 data file.
        output_name : str, optional
            Prefix of output files. Writes both pdf and dat file. 
        time_rate : bool, optional
            If True, calculates ct/keV/s. The defualt is ct/keV. 
        show_plot : bool, optional
            Wether or not to show plot (default is False).
        """

        # Log message:
        logger.info("getting raw spectrum...")

        # Option to read in binned data from hdf5 file:
        if binned_data:
            self.load_binned_data_from_hdf5(binned_data)
            self.get_binning_info() 

        # Option to normalize by total time:
        if time_rate==False:
            raw_rate = self.energy_hist/self.energy_bin_widths
            ylabel = "$\mathrm{ct \ keV^{-1}}$"
            data_label = "Rate[ct/keV]"
        if time_rate==True:
            raw_rate = self.energy_hist/self.energy_bin_widths/self.total_time
            ylabel = "$\mathrm{ct \ keV^{-1} \ s^{-1}}$"
            data_label = "Rate[ct/keV/s]"

        # Plot:
        plot_kwargs = {"label":"raw spectrum", "ls":"", "marker":"o", "color":"black"}
        fig_kwargs = {"xlabel":"Energy [keV]", "ylabel":ylabel}
        self.make_basic_plot(self.energy_bin_centers, raw_rate,\
            x_error=self.energy_bin_widths/2.0, output_name=output_name,\
            plot_kwargs=plot_kwargs, fig_kwargs=fig_kwargs, show_plots=show_plots)

        # Write data:
        if output_name != None:
            d = {"Energy[keV]":self.energy_bin_centers,data_label:raw_rate}
            df = pd.DataFrame(data=d)
            df.to_csv("%s.dat" %output_name,float_format='%10.5e',index=False,sep="\t",columns=["Energy[keV]",data_label])
        
        return

    def get_raw_lightcurve(self, binned_data=None, output_name=None, show_plots=False):

        """Calculates raw lightcurve of binned data, plots, and writes data to file.
        
        Parameters
        ----------
        binned_data : str, optional
            Name of binnned hdf5 data file to use.
        output_name : str, optional
            Prefix of output files. Writes both pdf and dat file. 
        show_plots : bool, optional
            Wether or not to show plot (default is False).
        """

        # Log message:
        logger.info("getting raw lightcurve...")

        # Option to read in binned data from hdf5 file:
        if binned_data:
            self.load_binned_data_from_hdf5(binned_data)
            self.get_binning_info() 
        
        # Calculate raw light curve:
        raw_lc = self.time_hist/self.time_bin_widths

        # Plot:
        plot_kwargs = {"ls":"-", "marker":"", "color":"black", "label":"raw lightcurve"}
        fig_kwargs = {"xlabel":"Time [s]", "ylabel":"Rate [$\mathrm{ct \ s^{-1}}$]"}
        self.make_basic_plot(self.time_bin_centers - self.time_bin_centers[0], raw_lc,\
            output_name=output_name, plt_scale="semilogy", 
            plot_kwargs=plot_kwargs, fig_kwargs=fig_kwargs, show_plots=show_plots)
            
        # Write data:
        if output_name != None:
            d = {"Time[UTC]":self.time_bin_centers,"Rate[ct/s]":self.time_hist/self.time_bin_widths}
            df = pd.DataFrame(data=d)
            df.to_csv("%s.dat" %output_name,index=False,sep="\t",columns=["Time[UTC]","Rate[ct/s]"])

        return



import os
import copy
import numpy as np
import astropy.units as u
import astropy.io.fits as fits
import logging
logger = logging.getLogger(__name__)

from histpy import Histogram

from .deconvolution_algorithm_base import DeconvolutionAlgorithmBase

class RichardsonLucy(DeconvolutionAlgorithmBase):
    """
    A class for the RichardsonLucy algorithm. 
    The algorithm here is based on Knoedlseder+99, Knoedlseder+05, Siegert+20.
    
    An example of parameter is as follows.

    iteration_max: 100
    minimum_flux:
        value: 0.0
        unit: "cm-2 s-1 sr-1"
    acceleration: True
    alpha_max: 10.0
    response_weighting: True 
    response_weighting_index: 0.5
    smoothing: True 
    smoothing_FWHM: 
      value: 2.0
      unit: "deg"
    background_normalization_optimization: True 
    background_normalization_range: {"albedo": [0.9, 1.1]}
    save_results: True
    save_results_directory: "./results"

    """

    def __init__(self, initial_model, dataset, mask, parameter):

        DeconvolutionAlgorithmBase.__init__(self, initial_model, dataset, mask, parameter)

        self.do_acceleration = parameter.get('acceleration', False)

        self.alpha_max = parameter.get('alpha_max', 1.0)

        self.do_response_weighting = parameter.get('response_weighting', False)
        if self.do_response_weighting:
            self.response_weighting_index = parameter.get('response_weighting_index', 0.5)

        self.do_smoothing = parameter.get('smoothing', False)
        if self.do_smoothing:
            self.smoothing_fwhm = parameter['smoothing_FWHM']['value'] * u.Unit(parameter['smoothing_FWHM']['unit'])
            logger.info(f"Gaussian filter with FWHM of {self.smoothing_fwhm} will be applied to delta images ...")

        self.do_bkg_norm_optimization = parameter.get('background_normalization_optimization', False)
        if self.do_bkg_norm_optimization:
            self.dict_bkg_norm_range = parameter.get('background_normalization_range', {key: [0.0, 100.0] for key in self.dict_bkg_norm.keys()})

        self.save_results = parameter.get('save_results', False)
        self.save_results_directory = parameter.get('save_results_directory', './results')

        if self.save_results is True:
            if os.path.isdir(self.save_results_directory):
                logger.warning(f"A directory {self.save_results_directory} already exists. Files in {self.save_results_directory} may be overwritten. Make sure that is not a problem.")
            else:
                os.makedirs(self.save_results_directory)

    def initialization(self):
        """
        initialization before running the image deconvolution
        """
        # clear counter 
        self.iteration_count = 0

        # clear results
        self.results.clear()

        # copy model
        self.model = copy.deepcopy(self.initial_model)

        # calculate exposure map
        self.summed_exposure_map = self.calc_summed_exposure_map()

        # mask setting
        if self.mask is None and np.any(self.summed_exposure_map.contents == 0):
            self.mask = Histogram(self.model.axes, contents = self.summed_exposure_map.contents > 0)
            self.model = self.model.mask_pixels(self.mask)
            logger.info("There are zero-exposure pixels. A mask to ignore them was set.")

        # response-weighting filter
        if self.do_response_weighting:
            self.response_weighting_filter = (self.summed_exposure_map.contents / np.max(self.summed_exposure_map.contents))**self.response_weighting_index
            logger.info("The response weighting filter was calculated.")

        # expected count histograms
        self.expectation_list = self.calc_expectation_list(model = self.initial_model, dict_bkg_norm = self.dict_bkg_norm)
        logger.info("The expected count histograms were calculated with the initial model map.")

        # calculate summed background models for M-step
        if self.do_bkg_norm_optimization:
            self.dict_summed_bkg_model = {}
            for key in self.dict_bkg_norm.keys():
                self.dict_summed_bkg_model[key] = self.calc_summed_bkg_model(key)

    def pre_processing(self):
        """
        pre-processing for each iteration
        """
        pass

    def Estep(self):
        """
        E-step (but it will be skipped).
        Note that self.expectation_list is updated in self.post_processing().
        """
        pass

    def Mstep(self):
        """
        M-step in RL algorithm.
        """

        ratio_list = [ data.event / expectation for data, expectation in zip(self.dataset, self.expectation_list) ]
        
        # delta model
        sum_T_product = self.calc_summed_T_product(ratio_list)
        self.delta_model = self.model * (sum_T_product/self.summed_exposure_map - 1)

        if self.mask is not None:
            self.delta_model = self.delta_model.mask_pixels(self.mask)
        
        # background normalization optimization
        if self.do_bkg_norm_optimization:
            for key in self.dict_bkg_norm.keys():

                sum_bkg_T_product = self.calc_summed_bkg_model_product(key, ratio_list)
                sum_bkg_model = self.dict_summed_bkg_model[key]
                bkg_norm = self.dict_bkg_norm[key] * (sum_bkg_T_product / sum_bkg_model)

                bkg_range = self.dict_bkg_norm_range[key]
                if bkg_norm < bkg_range[0]:
                    bkg_norm = bkg_range[0]
                elif bkg_norm > bkg_range[1]:
                    bkg_norm = bkg_range[1]

                self.dict_bkg_norm[key] = bkg_norm

    def post_processing(self):
        """
        Here three processes will be performed.
        - response weighting filter: the delta map is renormalized as pixels with large exposure times will have more feedback.
        - gaussian smoothing filter: the delta map is blurred with a Gaussian function.
        - acceleration of RL algirithm: the normalization of delta map is increased as long as the updated image has no non-negative components.
        """

        self.processed_delta_model = copy.deepcopy(self.delta_model)

        if self.do_response_weighting:
            self.processed_delta_model[:] *= self.response_weighting_filter

        if self.do_smoothing:
            self.processed_delta_model = self.processed_delta_model.smoothing(fwhm = self.smoothing_fwhm)
        
        if self.do_acceleration:
            self.alpha = self.calc_alpha(self.processed_delta_model, self.model)
        else:
            self.alpha = 1.0

        self.model = self.model + self.processed_delta_model * self.alpha
        self.model[:] = np.where(self.model.contents < self.minimum_flux, self.minimum_flux, self.model.contents)

        if self.mask is not None:
            self.model = self.model.mask_pixels(self.mask)
        
        # update expectation_list
        self.expectation_list = self.calc_expectation_list(self.model, dict_bkg_norm = self.dict_bkg_norm)
        logger.debug("The expected count histograms were updated with the new model map.")

        # update loglikelihood_list
        self.loglikelihood_list = self.calc_loglikelihood_list(self.expectation_list)
        logger.debug("The loglikelihood list was updated with the new expected count histograms.")

    def register_result(self):
        """
        The values below are stored at the end of each iteration.
        - iteration: iteration number
        - model: updated image
        - delta_model: delta map after M-step 
        - processed_delta_model: delta map after post-processing
        - alpha: acceleration parameter in RL algirithm
        - background_normalization: optimized background normalization
        - loglikelihood: log-likelihood
        """
        
        this_result = {"iteration": self.iteration_count, 
                       "model": copy.deepcopy(self.model), 
                       "delta_model": copy.deepcopy(self.delta_model),
                       "processed_delta_model": copy.deepcopy(self.processed_delta_model),
                       "background_normalization": copy.deepcopy(self.dict_bkg_norm),
                       "alpha": self.alpha, 
                       "loglikelihood": copy.deepcopy(self.loglikelihood_list)}

        # show intermediate results
        logger.info(f'  alpha: {this_result["alpha"]}')
        logger.info(f'  background_normalization: {this_result["background_normalization"]}')
        logger.info(f'  loglikelihood: {this_result["loglikelihood"]}')
        
        # register this_result in self.results
        self.results.append(this_result)

    def check_stopping_criteria(self):
        """
        If iteration_count is smaller than iteration_max, the iterative process will continue.

        Returns
        -------
        bool
        """
        if self.iteration_count < self.iteration_max:
            return False
        return True

    def finalization(self):
        """
        finalization after running the image deconvolution
        """
        if self.save_results == True:
            logger.info('Saving results in {self.save_results_directory}')

            # model
            for this_result in self.results:
                iteration_count = this_result["iteration"]

                this_result["model"].write(f"{self.save_results_directory}/model_itr{iteration_count}.hdf5", overwrite = True)
                this_result["delta_model"].write(f"{self.save_results_directory}/delta_model_itr{iteration_count}.hdf5", overwrite = True)
                this_result["processed_delta_model"].write(f"{self.save_results_directory}/processed_delta_model_itr{iteration_count}.hdf5", overwrite = True)

            #fits
            primary_hdu = fits.PrimaryHDU()

            col_iteration = fits.Column(name='iteration', array=[float(result['iteration']) for result in self.results], format='K')
            col_alpha = fits.Column(name='alpha', array=[float(result['alpha']) for result in self.results], format='D')
            cols_bkg_norm = [fits.Column(name=key, array=[float(result['background_normalization'][key]) for result in self.results], format='D') 
                             for key in self.dict_bkg_norm.keys()]
            cols_loglikelihood = [fits.Column(name=f"{self.dataset[i].name}", array=[float(result['loglikelihood'][i]) for result in self.results], format='D') 
                                  for i in range(len(self.dataset))]

            table_alpha = fits.BinTableHDU.from_columns([col_iteration, col_alpha])
            table_alpha.name = "alpha"

            table_bkg_norm = fits.BinTableHDU.from_columns([col_iteration] + cols_bkg_norm)
            table_bkg_norm.name = "bkg_norm"

            table_loglikelihood = fits.BinTableHDU.from_columns([col_iteration] + cols_loglikelihood)
            table_loglikelihood.name = "loglikelihood"

            hdul = fits.HDUList([primary_hdu, table_alpha, table_bkg_norm, table_loglikelihood])
            hdul.writeto(f'{self.save_results_directory}/results.fits',  overwrite=True)

    def calc_alpha(self, delta_model, model):
        """
        Calculate the acceleration parameter in RL algorithm.

        Returns
        -------
        float
            Acceleration parameter
        """
        diff = -1 * (model / delta_model).contents

        diff[(diff <= 0) | (delta_model.contents == 0)] = np.inf

        if self.mask is not None:
            diff[np.invert(self.mask.contents)] = np.inf

        alpha = min(np.min(diff), self.alpha_max)

        if alpha < 1.0:
            alpha = 1.0

        return alpha



from histpy import Histogram
from pathlib import Path
from cosipy.response import FullDetectorResponse
import h5py as h5
from histpy import Histogram, Axis, Axes
from cosipy.response import PointSourceResponse
import sys
from mhealpy import HealpixMap

class SourceInjector():
    
    def __init__(self, response_path, response_frame = "local"):

        """
        `SourceInjector` convolve response, source model(s) and orientation to produce a mocked simulated data. The data can be saved for data anlysis with cosipy.

        Parameters
        ----------
        response : str or pathlib.Path
            The path to the response file
        response_frame : str, optional
            The frame of the Compton data space (CDS) of the response. It only accepts `local` or "galactic". (the default is `local`, which means the CDS is in the local detector frame.
        """

        self.response_path =  response_path

        if response_frame == "local" or response_frame == "galactic":

            self.response_frame = response_frame

        else:
            raise ValueError("The response frame can only be `local` or `galactic`!")


    @staticmethod
    def get_psr_in_galactic(coordinate, response_path, spectrum):
        
        """
        Get the point source response (psr) in galactic. Please be aware that you must use a galactic response!
        To do: to make the weight parameter not hardcoded
        
        Parameters
        ----------
        coordinate : astropy.coordinates.SkyCoord
            The coordinate.
        response_path : str or path.lib.Path
            The path to the response.
        spectrum : astromodels.functions
            The spectrum of the source to be placed at the hypothesis coordinate.
        
        Returns
        -------
        psr : histpy.Histogram
            The point source response of the spectrum at the hypothesis coordinate.
        """
        
        # Open the response
        # Notes from Israel: Inside it contains a single histogram with all the regular axes for a Compton Data Space (CDS) analysis, in galactic coordinates. Since there is no class yet to handle it, this is how to read in the HDF5 manually.
        
        with h5.File(response_path) as f:

            axes_group = f['hist/axes']
            axes = []
            for axis in axes_group.values():
                # Get class. Backwards compatible with version
                # with only Axis
                axis_cls = Axis
                if '__class__' in axis.attrs:
                    class_module, class_name = axis.attrs['__class__']
                    axis_cls = getattr(sys.modules[class_module], class_name)
                axes += [axis_cls._open(axis)]
        axes = Axes(axes)
        
        # get the pixel number of the hypothesis coordinate
        map_temp = HealpixMap(base = axes[0])
        coordinate_pix_number = map_temp.ang2pix(coordinate)
        
        # get the expectation for the hypothesis coordinate (a point source)
        with h5.File(response_path) as f:
            pix = coordinate_pix_number
            psr = PointSourceResponse(axes[1:], f['hist/contents'][pix+1], unit = f['hist'].attrs['unit'])
                
        return psr


    def inject_point_source(self, spectrum, coordinate, orientation = None, source_name = "point_source",
                            make_spectrum_plot = False, data_save_path = None, project_axes = None):

        """
        Get the expected counts for a point source.

        Parameters
        ----------
        spectrum : astromodels.functions
            The spectrum model defined from `astromodels`.
        coordinate : astropy.coordinates.SkyCoord
            The coordinate of the point source.
        orientation : cosipy.spacecraftfile.SpacecraftFile, option
            The orientation of the telescope during the mock simulation. This is needed when using a detector response. (the default is `None`, which means a galactic response is used.
        source_name : str, optional
            The name of the source (the default is `point_source`).
        make_spectrum_plot : bool, optional
            Set `True` to make the plot of the injected spectrum.
        data_save_path : str or pathlib.Path, optional
            The path to save the injected data to a `.h5` file. This should include the file name. (the default is `None`, which means the injected data won't be saved.
        project_axes : list, optional
            The axes to project before saving the data file (the default is `None`, which means the data won't be projected).
            
        Returns
        -------
        histpy.Histogram
            The `Histogram object of the injected spectrum.`
        """
        
        
        # get the point source response in local frame
        if self.response_frame == "local":

            if orientation == None:
                raise TypeError("The when the data are binned in local frame, orientation must be provided to compute the expected counts.")
                
            # get the dwell time map
            coord_in_sc_frame = orientation.get_target_in_sc_frame(target_name = source_name, 
                                                                   target_coord = coordinate, 
                                                                   quiet = True)
            
            # get the dwell time map in the detector frame
            dwell_time_map = orientation.get_dwell_map(response = self.response_path)

            with FullDetectorResponse.open(self.response_path) as response:
                psr = response.get_point_source_response(dwell_time_map)

        # get the point source response in galactic frame
        elif self.response_frame == "galactic":

            psr = SourceInjector.get_psr_in_galactic(coordinate = coordinate, response_path = self.response_path, spectrum = spectrum)

        injected = psr.get_expectation(spectrum)
        # setting the Em and Ei scale to linear to match the simulated data
        # The linear scale of Em is the default for COSI data
        injected.axes["Em"].axis_scale = "linear"

        if project_axes is not None:
            injected = injected.project(project_axes)

        if make_spectrum_plot is True:
            ax, plot = injected.project("Em").draw(label = "Injected point source", color = "green")
            ax.legend()
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_ylabel("Counts")

        if data_save_path is not None:
            injected.write(data_save_path)

        return injected

        





# Imports:
import sys
import os
import yaml
import argparse
import cosipy.data_io
from yayc import Configurator

class DataIO:

    """Handles main inputs and outputs."""

    def __init__(self, input_yaml, pw=None):
    
        """
        Parameters
        ----------
        input_yaml : yaml file
            Input yaml file containing all needed inputs for analysis.
    
        Notes
        -----
        The main inputs must currently be passed with the yaml file.
        The parameter configurator will be updated in the near future,
        to allow for much more flexibility. 
        """
        
        # Data I/O:
        inputs = Configurator().open(input_yaml)
        self.data_file = inputs['data_file'] # Full path to input data file.
        self.ori_file = inputs['ori_file'] # Full path to ori file. 
        self.unbinned_output = inputs['unbinned_output'] # fits or hdf5
        self.time_bins = inputs['time_bins'] # Time bin size in seconds. Takes int, float, or list of bin edges.
        self.energy_bins = inputs['energy_bins'] # Needs to match response. Takes list. 
        self.phi_pix_size = inputs['phi_pix_size'] # Binning of Compton scattering angle [deg]
        self.nside = inputs['nside'] # Healpix binning of psi chi local
        self.scheme = inputs['scheme'] # Healpix binning of psi chi local
        self.tmin = inputs['tmin'] # Min time in seconds. 
        self.tmax = inputs['tmax'] # Max time in seconds. 
